"""Package defining guidance related functionality."""

from .agent_base import GuidanceAgent
from .agent_default import (
    DefaultGuidanceAgent,
    ArrowGuidanceActuator,
    BoxGuidanceActuator,
)
from .sensor_guidance import TaskAcceptabilitySensor
from .sensor_system_monitoring import SystemMonitoringTaskAcceptabilitySensor
from .sensor_tracking import TrackingTaskAcceptabilitySensor
from .sensor_resource_management import ResourceManagementTaskAcceptabilitySensor

__all__ = (
    "GuidanceAgent",
    "DefaultGuidanceAgent",
    "ArrowGuidanceActuator",
    "BoxGuidanceActuator",
    "TaskAcceptabilitySensor",
    "SystemMonitoringTaskAcceptabilitySensor",
    "TrackingTaskAcceptabilitySensor",
    "ResourceManagementTaskAcceptabilitySensor",
)



"""Module contains a default implementation for a guidance agent, see `DefaultGuidanceAgent` documentation for details."""

import random
from typing import Any, Literal
from collections.abc import Iterable
from star_ray.agent import Actuator, Sensor, observe
from icua.event import MouseMotionEvent, EyeMotionEvent
from icua.agent import (
    ArrowGuidanceActuator,
    BoxGuidanceActuator,
)
from icua.utils import LOGGER
from .agent_base import GuidanceAgent

__all__ = (
    "DefaultGuidanceAgent",
    "ArrowGuidanceActuator",
    "BoxGuidanceActuator",
)


class DefaultGuidanceAgent(GuidanceAgent):
    """Default implementation of a guidance agent for the matbii system.

    Show guidance if:
        1. there is no guidance already active.
        2. the task is unacceptable.
        3. the user is not already attending on the task.
        4. the grace period has elapsed.
    """

    # used to break ties when multiple tasks could be highlighted. See `break_tie` method.
    BREAK_TIES = ("random", "longest")

    # condition 4. - whether to track the time since the last failure, or since the last guidance was shown
    GRACE_ON = ("guidance_task", "guidance_any", "failure", "attention")

    # method to use to determine which task the user is attending to
    ATTENTION_MODES = ("fixation", "gaze", "mouse")

    # if we havent had fresh eyetracking data for more than this, there may be something wrong... display a warning
    MISSING_GAZE_DATA_THRESHOLD = 0.1

    def __init__(
        self,
        sensors: list[Sensor],
        actuators: list[Actuator],
        break_ties: Literal["random", "longest"] = "random",
        grace_mode: Literal[
            "guidance_task", "guidance_any", "failure", "attention"
        ] = "failure",
        attention_mode: Literal["fixation", "gaze", "mouse"] = "fixation",
        grace_period: float = 3.0,
        counter_factual: bool = False,
        **kwargs: dict[str, Any],
    ):
        """Constructor.

        Args:
            sensors (list[Sensor]): list of sensors, this will typically be a list of `icua.agent.TaskAcceptabilitySensor`s. A `UserInputSensor` will always be added automatically.
            actuators (list[Actuator]): list of actuators, this will typically contain actuators that are capable of providing visual feedback to a user, see e.g. `icua.agent.GuidanceActuator` and its concrete implementations. A `CounterFactualGuidanceActuator` will be added by default.
            break_ties (Literal["random", "longest", "since"], optional): how to break ties if guidance may be shown on multiple tasks simultaneously. Defaults to "random". "random" will randomly break the tie, "longest" will choose the task that has been in failure for the longest, "since" will choose the task that has not had guidance shown for the longest time.
            grace_mode (Literal["guidance_task", "guidance_any", "failure", "attention"], optional): how to track the grace period. "guidance_task" will track the time since guidance was last shown on the task, "guidance_any" will track the time since guidance was last shown on any task, "failure" will track the time since the last failure on the task, "attention" will track the time since the user was last attending to a task. Defaults to "failure".
            attention_mode (Literal["fixation", "gaze", "mouse"], optional): method of determining where the user is attending. "fixation" will use the most recent gaze fixation, "gaze" will use the gaze position (including saccades), "mouse" will use the current mouse position. Defaults to "fixation".
            grace_period (float, optional): the time to wait (seconds) before guidance may be shown for a task after the last time guidance was shown on the task. Defaults to 3.0 seconds.
            counter_factual (bool, optional): whether guidance should be shown to the user, or whether it should just be logged.  This allows counter-factual experiments to be run, we can track when guidance would have been shown, and compare the when it was actually shown (in a different run). Defaults to False.
            kwargs (dict[str,Any]): Additional optional keyword arguments.
        """
        super().__init__(
            list(filter(None, sensors)),
            list(filter(None, actuators)),
            counter_factual=counter_factual,
            **kwargs,
        )
        self._attention_mode = attention_mode
        if self._attention_mode not in DefaultGuidanceAgent.ATTENTION_MODES:
            raise ValueError(
                f"Invalid argument: `attention_mode` {self._attention_mode} must be one of {DefaultGuidanceAgent.ATTENTION_MODES}"
            )

        self._grace_mode = grace_mode
        if self._grace_mode not in DefaultGuidanceAgent.GRACE_ON:
            raise ValueError(
                f"Invalid argument: `grace_mode` {self._grace_mode} must be one of {DefaultGuidanceAgent.GRACE_ON}"
            )

        self._break_ties = break_ties
        if self._break_ties not in DefaultGuidanceAgent.BREAK_TIES:
            raise ValueError(
                f"Invalid argument: `break_ties` {self._break_ties} must be one of {DefaultGuidanceAgent.BREAK_TIES}"
            )

        self._grace_period = grace_period  # TODO
        if self._grace_period < 0.0:
            raise ValueError(
                f"Invalid argument: `grace_period` {self._grace_period} must be > 0 "
            )

        # used to track the tasks that the user is currently attending to
        self._attending_tasks = set()

    def on_attending(self, attending_tasks: set[str]) -> None:
        """Called when the user's attention changes.

        Args:
            attending_tasks (set[str]): the new set of tasks that the user is currently attending to.
        """
        for task in attending_tasks:
            self.beliefs[task]["last_attended"] = self.get_cycle_start()
            self._log_info(task, "attending", True)
        if len(attending_tasks) == 0:
            self._log_info("none", "attending", True)

    def decide(self):  # noqa
        # update when the user was last attending to a task
        attending_tasks = self.get_attending_tasks()  # empty or a single task
        if attending_tasks != self._attending_tasks:
            self._attending_tasks = attending_tasks
            self.on_attending(attending_tasks)

        # make guidance decisions
        if self.guidance_on_tasks:
            # guidance is active, should it be?
            guidance_and_acceptable = self.guidance_on_tasks & self.acceptable_tasks
            if guidance_and_acceptable:
                # the task has guidance showing, but is now acceptable - hide guidance for this task
                for task in guidance_and_acceptable:
                    self.hide_guidance(task)
                return

            guidance_and_attending = self.guidance_on_tasks & self._attending_tasks
            if guidance_and_attending:
                # the task has guidance showing, but the user is attending to it - hide guidance for this task
                for task in guidance_and_attending:
                    self.hide_guidance(task)
                return

            # keep showing guidance
        else:
            # guidance is NOT active, should it be?
            # these tasks are candidates for showing guidance
            unacceptable = self.unacceptable_tasks
            # remove those tasks that the user is currently attending
            unacceptable -= self._attending_tasks
            # check the grace period
            unacceptable = self.grace_period_over(unacceptable)

            task = self.break_tie(unacceptable)
            if task:
                # we have selected a task to show guidance on
                self.show_guidance(task)

            # no task meets the criteria, the user is doing well!

    def grace_period_over(self, tasks: set[str]) -> set[str]:
        """Get the set of (unacceptable) tasks that have had their grace period elapse.

        Args:
            tasks (set[str]): the set of tasks to check.

        Returns:
            set[str]: the set of tasks that have had their grace period elapse.
        """
        if self._grace_mode == "guidance_task":
            # how long since guidance was last shown on the task?
            def _grace_met(t):
                tsgs = self.time_since_guidance_start(t)
                # check if grace period has elapsed, or if guidance has never been shown on the task (NaN value)
                return tsgs > self._grace_period or tsgs != tsgs

            return set(t for t in tasks if _grace_met(t))
        elif self._grace_mode == "guidance_any":
            # how long since guidance was last shown on any task?
            tsgs = self.time_since_guidance_start(None)
            if tsgs > self._grace_period or tsgs != tsgs:
                return tasks  # guidance has not been shown recently
            else:
                return set()  # a task has recently had guidance shown, we should not show guidance for any task yet
        elif self._grace_mode == "attention":

            def _grace_met(t):
                tsgs = self.time_since_last_attended(t)
                # check if grace period has elapsed, or if guidance has never been shown on the task (NaN value)
                return tsgs > self._grace_period or tsgs != tsgs

            # print({t: (_grace_met(t), self.time_since_last_attended(t)) for t in tasks})
            # how long since the user was last attending to a task?
            return set(t for t in tasks if _grace_met(t))
        elif self._grace_mode == "failure":
            # how long has this task been in failure?
            return set(
                t
                for t in tasks
                if self.time_since_failure_start(t) > self._grace_period
            )
        else:
            raise ValueError(
                f"Unknown grace mode: {self._grace_mode}, must be one of {DefaultGuidanceAgent.GRACE_ON}"
            )

    def time_since_last_attended(self, task: str) -> float:
        """Get the time since the user was last attending to a task."""
        return self.get_cycle_start() - self.beliefs[task].get(
            "last_attended", float("nan")
        )

    def get_attending_tasks(self) -> set[str]:
        """Get the set of tasks that the user is currently attending to - this is typically only one task.

        Returns:
            set[str]: set of tasks that the user is currently attending to (empty if no tasks are being attended to).
        """
        if self._attention_mode == "fixation":
            result = self.attending_task_fixation()
        elif self._attention_mode == "gaze":
            result = self.attending_task_gaze()
        elif self._attention_mode == "mouse":
            result = self.attending_task_mouse()
        else:
            raise ValueError(f"Unknown attention mode: {self._attention_mode}")
        if result is None:
            return set()
        else:
            return set([result])

    def attending_task_mouse(self) -> str | None:
        """Use mouse position to determine which task the user is attending to."""
        return self.mouse_target

    def attending_task_gaze(self) -> str | None:
        """Use gaze position (including saccades) to determine which task the user is attending to."""
        return self.gaze_target

    def attending_task_fixation(self) -> str | None:
        """Use fixation to determine which task the user is attending to."""
        #print(self.fixation_target)
        return self.fixation_target

    def break_tie(
        self,
        tasks: Iterable[str],
        method: Literal["random", "longest"] = "random",
    ) -> str | None:
        """Break a tie on tasks that all met the criteria for displaying guiance.

        The tie is broken using the `method` argument:
        - "random" will randomly break the tie.
        - "longest" will choose the task that has been in failure for the longest.

        Args:
            tasks (Iterable[str]): tasks to break the tie, one of which will be returned.
            method (Literal["random", "longest"], optional): how to break ties if guidance may be shown on multiple tasks simultaneously. Defaults to "random".

        Returns:
            str: the chosen task, or None if there are no tasks to choose from.
        """
        if len(tasks) == 0:
            return None
        if method == "random":
            # randomly break the tie
            return random.choice(list(tasks))
        elif method == "longest":
            # choose the task longest in failure
            return max(
                [(task, self.time_since_failure_start(task)) for task in tasks],
                key=lambda x: x[1],
            )[0]
        else:
            raise ValueError(
                f"Unknown guidance tie break method: {method}, must be one of {DefaultGuidanceAgent.BREAK_TIES}"
            )

    @observe([EyeMotionEvent, MouseMotionEvent])
    def _on_motion_event(self, event: MouseMotionEvent | EyeMotionEvent):
        """It may be useful to the actuators to get these events. It is a trick to forward sensory input to the agents actuators. The `ArrowGuidanceActuator` is an example that requires this information to display the array at the gaze/mouse position."""
        # TODO we may need to guard against actuators executing these actions...?
        # manually attempt the event, we could specify which actuators need this information...?
        self.attempt(event)

    def on_acceptable(self, task: str):  # noqa
        self._log_info(task, "acceptable", True)
        return super().on_acceptable(task)

    def on_unacceptable(self, task: str):  # noqa
        self._log_info(task, "acceptable", False)
        return super().on_unacceptable(task)

    def _log_info(self, task, z, ok):
        """Log info related to a task to the console."""
        info = "task %20s %20s %s" % (z, task, ["✘", "✔"][int(ok)])
        LOGGER.info(info)



"""Entry point of the `matbii` simulation.

`matbii` may be run with `python -m matbii -c <CONFIG_FILE>` after installing with pip, where <CONFIG_FILE> is a path to a json configuration file, see [documentation](https://dicelab-rhul.github.io/matbii/latest/) for details on configuring `matbii`.
"""

from functools import partial
from typing import Any


# imports for creating the environment
from matbii.environment import MultiTaskEnvironment

# imports for creating guidance agents
from matbii.guidance import (
    DefaultGuidanceAgent,
    SystemMonitoringTaskAcceptabilitySensor,
    TrackingTaskAcceptabilitySensor,
    ResourceManagementTaskAcceptabilitySensor,
)
from matbii.agent import (
    TrackingActuator,
    SystemMonitoringActuator,
    ResourceManagementActuator,
)
from matbii.avatar import (
    Avatar,
    AvatarActuator,
    ExitActuator,
    AvatarTrackingActuator,
    AvatarSystemMonitoringActuator,
    AvatarResourceManagementActuator,
)
from matbii.config import Configuration
from matbii.utils import (
    TASK_PATHS,
    TASK_ID_TRACKING,
    TASK_ID_RESOURCE_MANAGEMENT,
    TASK_ID_SYSTEM_MONITORING,
)


def main(
    config: str | None = None,
    **context: dict[str, Any],
):
    """Run the `matbii` simulation."""
    # args from command line can be used in config
    # context = dict(
    #     experiment=dict(id=experiment) if experiment else dict(),
    #     participant=dict(id=participant) if participant else dict(),
    #     **kwargs,
    # )

    # load configuration from the config file
    config = Configuration.from_file(config, context=context)
    # initialise logging
    config = Configuration.initialise_logging(config)

    # Create the avatar:
    # - required sensors are added by default
    # - task related actuators are added when their corresponding task is enabled
    avatar = Avatar(
        [],
        [AvatarActuator(), ExitActuator()],  # will log user events by default
        window_config=config.window,
    )

    # if eyetracking is enabled, add a sensor to the avatar
    eyetracking_sensor = config.eyetracking.new_eyetracking_sensor()
    if eyetracking_sensor:
        avatar.add_component(eyetracking_sensor)

    agents = []  # will be given to the environment

    # create the guidance agent
    # - sensors will determine the acceptability of each task - if the task is enabled.
    # - change actuators for different guidance to be shown (must inherit from GuidanceActuator)
    guidance_agent = DefaultGuidanceAgent(
        [
            # add more if there are more tasks!
            SystemMonitoringTaskAcceptabilitySensor(),
            ResourceManagementTaskAcceptabilitySensor(),
            TrackingTaskAcceptabilitySensor(),
        ],
        [
            # used to log this agents beliefs for post experiment analysis
            # LogActuator(path=Path(config.logging.path) / "guidance_logs.log"),
            # shows arrow pointing at a task as guidance
            config.guidance.arrow.to_actuator(),
            # shows a box around a task as guidance
            config.guidance.box.to_actuator(),
        ],
        break_ties=config.guidance.break_ties,
        grace_period=config.guidance.grace_period,
        grace_mode=config.guidance.grace_mode,
        attention_mode=config.guidance.attention_mode,
        counter_factual=config.guidance.counter_factual,
    )
    agents.append(guidance_agent)

    env = MultiTaskEnvironment(
        wait=0.01,  # this can be zero as long as it doesnt matter that the env scheduler hogs asyncio: TODO test this with IO devices (eyetracker particularly)
        avatar=avatar,
        agents=agents,
        svg_size=(config.ui.width, config.ui.height),
        logging_path=config.logging.path,
        terminate_after=config.experiment.duration,
    )

    # NOTE: if you have more tasks to add, add them here!
    env.add_task(
        name=TASK_ID_TRACKING,
        path=[config.experiment.path, TASK_PATHS[TASK_ID_TRACKING]],
        agent_actuators=[TrackingActuator],
        avatar_actuators=[
            partial(
                AvatarTrackingActuator,
                # Negative values will invert the direction of movement
                target_speed=-50.0,  # TODO a config option for this?
            )
        ],
        enable=TASK_ID_TRACKING in config.experiment.enable_tasks,
    )

    env.add_task(
        name=TASK_ID_SYSTEM_MONITORING,
        path=[config.experiment.path, TASK_PATHS[TASK_ID_SYSTEM_MONITORING]],
        agent_actuators=[SystemMonitoringActuator],
        avatar_actuators=[AvatarSystemMonitoringActuator],
        enable=TASK_ID_SYSTEM_MONITORING in config.experiment.enable_tasks,
    )

    env.add_task(
        name=TASK_ID_RESOURCE_MANAGEMENT,
        path=[config.experiment.path, TASK_PATHS[TASK_ID_RESOURCE_MANAGEMENT]],
        agent_actuators=[ResourceManagementActuator],
        avatar_actuators=[AvatarResourceManagementActuator],
        enable=TASK_ID_RESOURCE_MANAGEMENT in config.experiment.enable_tasks,
    )
    env.run()



"""Module that implements the "system monitoring" task.

This files contains:
    - avatar actuator: `AvatarSystemMonitoringActuator`
    - agent actuator: `SystemMonitoringActuator`
    - actions: [`TargetMoveAction`, `PerturbSliderAction`, `ResetSliderAction`, `SetSliderAction`, `SetLightAction`, `ToggleLightAction`]
"""

import random
import re
from typing import Any, ClassVar, Union
from functools import partial
from pydantic import Field, field_validator


from icua.agent import agent_actuator, attempt, Actuator
from icua.event import XMLUpdateQuery, MouseButtonEvent

from star_ray_xml import (
    XMLState,
    Expr,
    update,
    select,
)

# these are constants that reflect the task svg TODO move to _const?
VALID_LIGHT_IDS = [1, 2]
VALID_SLIDER_IDS = [1, 2, 3, 4]


__all__ = (
    "AvatarSystemMonitoringActuator",
    "SystemMonitoringActuator",
    "PerturbSliderAction",
    "ResetSliderAction",
    "SetSliderAction",
    "SetLightAction",
    "ToggleLightAction",
)


class AvatarSystemMonitoringActuator(Actuator):
    """Actuator class that should be added to the Avatar when the system monitoring task is enabled. It allows the user to control the lights and sliders in this task."""

    def __init__(self, *args: list[Any], **kwargs: dict[str, Any]):
        """Constructor.

        Args:
            args (list[Any], optional): additional optional arguments.
            kwargs (dict[Any], optional): additional optional keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self._get_light_targets = partial(
            AvatarSystemMonitoringActuator.get_click_targets, r"light-(\d+)-button"
        )
        self._get_slider_targets = partial(
            AvatarSystemMonitoringActuator.get_click_targets, r"slider-(\d+)-button"
        )

    @attempt
    def attempt_mouse_event(
        self, user_action: MouseButtonEvent
    ) -> list[Union["SetLightAction", "SetSliderAction"]]:
        """Attempt method that takes a `MouseButtonEvent` as input from the user. It is up to the avatar to provide this event to this actuator. The information is essential if the user is to control this task.

        This will attempt either a `SetLightAction` or a `SetSliderAction` depending on what svg element was clicked (assuming the mouse event registers as a click).

        Effects:
        - `SetLightAction`: will always set the light to its preferred (acceptable) state (which is ON for light 1 and OFF for light 2).
        - `SetSliderAction` will reset the slider to its preferred (acceptable) state (which is the central position), see `ResetSliderAction`.

        Args:
            user_action (MouseButtonEvent): the users mouse button action.

        Returns:
            list[SetLightAction | SetSliderAction]: the action to be attempted
        """
        assert isinstance(user_action, MouseButtonEvent)
        actions = []
        if (
            user_action.status == MouseButtonEvent.DOWN
            and user_action.button == MouseButtonEvent.BUTTON_LEFT
        ):
            actions.extend(self._get_light_actions(user_action))
            actions.extend(self._get_slider_actions(user_action))
        return actions

    def _get_light_actions(
        self, user_action: MouseButtonEvent
    ) -> list["ToggleLightAction"]:
        targets = [int(x) for x in self._get_light_targets(user_action.target)]
        # the user can only set the state to the "acceptable" state
        acceptable = [
            SetLightAction.ON,
            SetLightAction.OFF,
        ]
        return [
            SetLightAction(target=target, state=acceptable[target - 1])
            for target in targets
        ]

    def _get_slider_actions(
        self, user_action: MouseButtonEvent
    ) -> list["SetSliderAction"]:
        targets = [int(x) for x in self._get_slider_targets(user_action.target)]
        return [ResetSliderAction(target=target) for target in targets]

    @staticmethod
    def get_click_targets(pattern: str, targets: list[str]) -> list[str]:
        """Extras the `id` field of a target in `targets` if it matches the pattern.

        Args:
            pattern (str): pattern to test.
            targets (list[str]): targets to search.

        Returns:
            list[str]: matching targets.
        """

        def _get():
            for target in targets:
                match = re.match(pattern, target)
                if match:
                    target = match.group(1)
                    yield target

        return list(_get())


@agent_actuator
class SystemMonitoringActuator(Actuator):
    """Actuator class that will be part of a `ScheduledAgent` or any other agent that controls the evolution of the system monitoring task."""

    @attempt
    def on_light(self, target: int) -> "SetLightAction":
        """Switch the `target` light to the "on" state.

        Args:
            target (int): the integer `id` of the target light (1 or 2).

        Returns:
            SetLightAction: the action
        """
        return SetLightAction(target=target, state=SetLightAction.ON)

    @attempt
    def off_light(self, target: int) -> "SetLightAction":
        """Switch the `target` light to the "off" state.

        Args:
            target (int): the integer `id` of the target light (1 or 2).

        Returns:
            SetLightAction: the action
        """
        return SetLightAction(target=target, state=SetLightAction.OFF)

    @attempt
    def toggle_light(self, target: int) -> "SetLightAction":
        """Toggle the `target` light (on->off, off->on).

        Args:
            target (int): the integer `id` of the target light (1 or 2).

        Returns:
            SetLightAction: the action
        """
        return ToggleLightAction(target=target)

    @attempt
    def perturb_slider(self, target: int) -> "SetSliderAction":
        """Perturb the `target` slider by +/- 1 slot.

        Args:
            target (int): the integer `id` of the target slider (1, 2, 3 or 4).

        Returns:
            SetSliderAction: the action
        """
        state = 2 * random.randint(0, 1) - 1  # randomly perturb +/- 1
        return SetSliderAction(target=target, state=state, relative=True)


def PerturbSliderAction(target: int) -> "SetSliderAction":
    """Perturb the `target` slider by +/- 1 slot.

    Args:
        target (int): the integer `id` of the target slider (1, 2, 3 or 4).

    Returns:
        SetSliderAction: the action
    """
    state = random.randint(0, 1) * 2 - 1
    return SetSliderAction(target=target, state=state, relative=True)


def ResetSliderAction(target: int) -> "SetSliderAction":
    """Reset the `target` slider to its central position.

    Args:
        target (int): the integer `id` of the target slider (1, 2, 3 or 4).

    Returns:
        SetSliderAction: the action
    """
    return SetSliderAction(target=target, state=None, relative=False)


class SetSliderAction(XMLUpdateQuery):
    """Action class that will update a sliders's state. The state of a slider is the integer index of one of its increments."""

    # slider to target
    target: int
    # state to set, or offset from current state, or None if should reset the state.
    state: int | None
    # is `state` relative to the current state?
    relative: bool = Field(default_factory=lambda: False)

    @field_validator("target", mode="before")
    @classmethod
    def _validate_target(cls, value):
        if value not in VALID_SLIDER_IDS:
            raise ValueError(f"`target` {value} must be one of {VALID_LIGHT_IDS}")
        return value

    @staticmethod
    def acceptable_state(increments: int) -> int:
        """Defines the acceptable state of the slider. This is the state that the slider will return to if the actions `state` field is None.

        Args:
            increments (int): the number of increments in the slider.

        Returns:
            int: the acceptable state of the slider
        """
        return increments // 2

    def __execute__(self, xml_state: XMLState) -> Any:  # noqa
        # get min and max values for the number of increments
        inc_target = f"slider-{self.target}-incs"
        but_target = f"slider-{self.target}-button"
        response = xml_state.select(
            select(
                xpath=f"//*[@id='{inc_target}']/svg:line",
                attrs=["y1", "data-state"],
            )
        )
        states = {x["data-state"]: x["y1"] for x in response}
        # TODO check that these are all the same?
        inc_size = states[2] - states[1]
        min_state, max_state = (min(states.keys()), max(states.keys()) - 1)
        state = self.state
        if state is None:
            self.relative = False
            state = SetSliderAction.acceptable_state(max_state + 1)

        # we select the parent of the button node because it contains the state and position to update
        xpath_parent = f"//*[@id='{but_target}']/parent::node()"
        if self.relative:
            # update the state relative to the current state
            response = xml_state.select(
                select(
                    xpath=xpath_parent,
                    attrs=["data-state"],
                )
            )
            assert len(response) == 1
            state = response[0]["data-state"] + self.state

        # new state should not overflow
        state = min(max(min_state, state), max_state)
        new_y = states[state] - inc_size
        return xml_state.update(
            update(
                xpath=xpath_parent,
                attrs={"data-state": state, "y": new_y},
            )
        )


class SetLightAction(XMLUpdateQuery):
    """Action class that will update a light's state (on=1 or off=0)."""

    target: int  # the target light
    state: int  # the new state of the light

    OFF: ClassVar[int] = 0  # off state
    ON: ClassVar[int] = 1  # on state

    @field_validator("target", mode="before")
    @classmethod
    def _validate_target(cls, value):
        if value not in VALID_LIGHT_IDS:
            raise ValueError(f"`target` {value} must be one of {VALID_LIGHT_IDS}")
        return value

    @field_validator("state", mode="before")
    @classmethod
    def _validate_state(cls, value: int | str):
        if isinstance(value, str):
            value = SetLightAction.coerce_light_state(value)
        if value not in (SetLightAction.OFF, SetLightAction.ON):
            raise ValueError(
                f"Invalid state `{value}` must be one of {[SetLightAction.OFF, SetLightAction.ON]}"
            )
        return value

    @staticmethod
    def coerce_light_state(value: str | int) -> int:
        """Coerce a state that is specified as a string to an int: "off -> 0, "on" -> 1.

        Args:
            value (str | int): the value to coerce valid values: 0,1,"on","off.

        Raises:
            ValueError: if the given state cannot be  coerced.

        Returns:
            int: the integer state.
        """
        if isinstance(value, int):
            return value
        if value == "on":
            return SetLightAction.ON
        elif value == "off":
            return SetLightAction.OFF
        else:
            raise ValueError(f"Invalid state `{value}` must be one of ['on', 'off']")

    def __execute__(self, xml_state: XMLState):  # noqa
        xml_state.update(
            update(
                xpath=f"//*[@id='light-{self.target}-button']",
                attrs={
                    "data-state": str(self.state),
                    "fill": Expr("{data-colors}[{state}]", state=self.state),
                },
            )
        )


class ToggleLightAction(XMLUpdateQuery):
    """Action class that will toggle a light's state from on->off and off->on."""

    target: int

    @field_validator("target", mode="before")
    @classmethod
    def _validate_target(cls, value):
        if value not in VALID_LIGHT_IDS:
            raise ValueError(f"`target` {value} must be one of {VALID_LIGHT_IDS}")
        return value

    def __execute__(self, xml_state: XMLState):  # noqa
        xml_state.update(
            update(
                xpath=f"//*[@id='light-{self.target}-button']",
                attrs={
                    "data-state": Expr("1-{data-state}"),
                    # GOTCHA! data-state will be updated first (above) and used to update fill! the dict order matters here.
                    "fill": Expr("{data-colors}[{data-state}]"),
                },
            )
        )



"""Module that implements the [resource management task][matbii.tasks.resource_management].

This files contains:
    - avatar actuator: `AvatarResourceManagementActuator`
    - agent actuator: `ResourceManagementActuator`
    - actions: [`SetPumpAction`, `TogglePumpAction`, `TogglePumpFailureAction`, `PumpFuelAction`, `BurnFuelAction`]
"""

import re
from typing import ClassVar, Literal, Any
from pydantic import field_validator
from functools import partial
from star_ray_xml import update, select, XMLState, Expr
from star_ray_xml.query import XMLUpdateQuery

from icua.event import MouseButtonEvent
from icua.agent import attempt, Actuator

TANK_IDS = list("abcdef")
TANK_MAIN_IDS = list("ab")
TANK_INF_IDS = list("ef")
PUMP_IDS = ["ab", "ba", "ca", "ec", "ea", "db", "fd", "fb"]
ALL = "*"

__all__ = (
    "ResourceManagementActuator",
    "AvatarResourceManagementActuator",
    "SetPumpAction",
    "TogglePumpAction",
    "BurnFuelAction",
    "PumpFuelAction",
    "TogglePumpFailureAction",
)


class AvatarResourceManagementActuator(Actuator):
    """Actuator class that should be added to the Avatar when the resource management task is enabled. It allows the user to control the pumps in this task.."""

    def __init__(self, *args: list[Any], **kwargs: dict[str, Any]):
        """Constructor.

        Args:
            args (list[Any], optional): additional optional arguments.
            kwargs (dict[Any], optional): additional optional keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self._get_pump_targets = partial(
            AvatarResourceManagementActuator.get_click_targets, r"pump-([a-z]+)-button"
        )

    @attempt
    def attempt_mouse_event(
        self, user_action: MouseButtonEvent
    ) -> list["TogglePumpAction"]:
        """Attempt method that takes a `MouseButtonEvent` as input from the user. It is up to the avatar to provide this event to this actuator. The information is essential if the user is to control this task.

        This will attempt a `TogglePumpAction` depending on what svg element was clicked (assuming the mouse event registers as a click).

        Effects:
        - `TogglePumpAction`: toggle the pump on->off or off->on.

        Args:
            user_action (MouseButtonEvent): the users mouse button action.

        Returns:
            list[TogglePumpAction]: the action to be attempted
        """
        assert isinstance(user_action, MouseButtonEvent)
        # always include the user action as it needs to be logged
        actions = []
        if (
            user_action.status == MouseButtonEvent.DOWN
            and user_action.button == MouseButtonEvent.BUTTON_LEFT
        ):
            actions.extend(self._get_pump_actions(user_action))
        return actions

    def _get_pump_actions(
        self, user_action: MouseButtonEvent
    ) -> list["TogglePumpAction"]:
        targets = self._get_pump_targets(user_action.target)
        return [TogglePumpAction(target=target) for target in targets]

    @staticmethod
    def get_click_targets(pattern: str, targets: list[str]) -> list[str]:
        """Extras the `id` field of a target in `targets` if it matches the pattern.

        Args:
            pattern (str): pattern to test.
            targets (list[str]): targets to search.

        Returns:
            list[str]: matching targets.
        """

        def _get():
            for target in targets:
                match = re.match(pattern, target)
                if match:
                    target = match.group(1)
                    yield target

        return list(_get())


class ResourceManagementActuator(Actuator):
    """Actuator class that will be part of a `ScheduledAgent` or any other agent that controls the evolution of the resource management task."""

    @attempt
    def burn_fuel(
        self,
        target: int | str,
        burn: float,
    ) -> "BurnFuelAction":
        """Burns a given amount of fuel in the `target` tank, if the tank is empty this has no effect.

        Args:
            target (int | Literal["a", "b"]): target tank
            burn (float): amount of fuel to burn.

        Returns:
            BurnFuelAction: the action
        """
        return BurnFuelAction(target=target, burn=burn)

    @attempt
    def pump_fuel(
        self,
        target: int | str,
        flow: float,
    ) -> "PumpFuelAction":
        """Pumps the given amount of fuel via the given pump.

        Args:
            target (int | Literal["ab", "ba", "ca", "ec", "ea", "db", "fd", "fb"]): target pump (this will determine which tanks are pumped to/from).
            flow (float): amount of fuel to pump.

        Returns:
            PumpFuelAction: the action
        """
        return PumpFuelAction(target=target, flow=flow)

    @attempt
    def toggle_pump_failure(
        self,
        target: int | str,
    ) -> "TogglePumpFailureAction":
        """Toggle a pump failure (on/off -> failure, failure -> off).

        Args:
            target (int | Literal["ab", "ba", "ca", "ec", "ea", "db", "fd", "fb"]): target pump.

        Returns:
            TogglePumpFailureAction: the action
        """
        return TogglePumpFailureAction(target=target)

    @attempt
    def toggle_pump(
        self,
        target: int | str,
    ) -> "TogglePumpAction":
        """Toggle pump state (on -> off, off -> on).

        Args:
            target (int | Literal["ab", "ba", "ca", "ec", "ea", "db", "fd", "fb"]): target pump.

        Returns:
            TogglePumpFailureAction: the action
        """
        return TogglePumpAction(target=target)


class PumpAction(XMLUpdateQuery):
    """Base class for pump related actions."""

    target: str

    OFF: ClassVar[int] = 0
    ON: ClassVar[int] = 1
    FAILURE: ClassVar[int] = 2

    @field_validator("target", mode="before")
    @classmethod
    def _validate_target(cls, value: str | int):
        if isinstance(value, int):
            value = PUMP_IDS[value]
        if value not in PUMP_IDS:
            raise ValueError(
                f"{SetPumpAction.__name__} `target` {value} must be one of {PUMP_IDS}"
            )
        return value


class SetPumpAction(PumpAction):
    """Action class for setting the state of a pump. Pumps states are off=0, on=1, failure=2."""

    state: int

    @field_validator("state", mode="before")
    @classmethod
    def _validate_state(cls, value: str | int):
        if isinstance(value, str):
            value = SetPumpAction.coerce_pump_state(value)
        if value not in (PumpAction.OFF, PumpAction.ON, PumpAction.FAILURE):
            raise ValueError(
                f"Invalid state `{value}` must be one of {[PumpAction.OFF, PumpAction.ON, PumpAction.FAILURE]}"
            )
        return value

    @staticmethod
    def coerce_pump_state(value: int | str) -> int:
        """Coerce a state that is specified as a string to an int: "off -> 0, "on" -> 1, "failure" -> 2.

        Args:
            value (str | int): the value to coerce.

        Raises:
            ValueError: if the given state cannot be coerced.

        Returns:
            int: the integer state.
        """
        if isinstance(value, int):
            return value
        if value == "on":
            return PumpAction.ON
        elif value == "off":
            return PumpAction.OFF
        elif value == "failure":
            return PumpAction.FAILURE
        else:
            raise ValueError(
                f"Invalid state `{value}` must be one of ['on', 'off', 'failure']"
            )

    @staticmethod
    def new_on(target: int | str) -> "SetPumpAction":
        """Factory method that returns an action that will turn the given pump on.

        Args:
            target (int | str): target pump.

        Returns:
            SetPumpAction: the action
        """
        return SetPumpAction(target=target, state=PumpAction.ON)

    @staticmethod
    def new_off(target: int | str) -> "SetPumpAction":
        """Factory method that returns an action that will turn the given pump off.

        Args:
            target (int | str): target pump.

        Returns:
            SetPumpAction: the action
        """
        return SetPumpAction(target=target, state=PumpAction.OFF)

    @staticmethod
    def new_failure(target: int | str) -> "SetPumpAction":
        """Factory function that returns an action that will make the pump fail.

        Args:
            target (int | str): target pump.

        Returns:
            SetPumpAction: the action
        """
        return SetPumpAction(target=target, state=PumpAction.FAILURE)

    def __execute__(self, state: XMLState):  # noqa
        return state.update(
            update(
                xpath=f"//*[@id='pump-{self.target}-button']",
                attrs={
                    "data-state": str(self.state),
                    "fill": Expr("{data-colors}[{state}]", state=self.state),
                },
            )
        )


class TogglePumpAction(PumpAction):
    """Action class that will toggle the pump on -> off, off -> on."""

    @staticmethod
    def new(
        target: int | Literal["ab", "ba", "ca", "ec", "ea", "db", "fd", "fb"],
    ) -> "TogglePumpAction":
        """Factory function for a `TogglePumpAction`.

        Args:
            target (int | Literal["ab", "ba", "ca", "ec", "ea", "db", "fd", "fb"]): target pump

        Returns:
            TogglePumpAction: the action.
        """
        return TogglePumpAction(target=target)

    def __execute__(self, xml_state: XMLState):  # noqa
        # check if pump is in a failure state
        pump_id = f"pump-{self.target}-button"
        # 0 -> 1, 1 -> 0, 2 -> 2 (cannot toggle if the pump is in failure)
        new_state = "(1-{data-state})%3"
        xml_state.update(
            update(
                xpath=f"//*[@id='{pump_id}']",
                attrs={
                    "data-state": Expr(new_state),
                    # GOTCHA! data-state will be updated first (from above) and used to update fill! the order matters here.
                    "fill": Expr("{data-colors}[{data-state}]"),
                },
            )
        )


class TogglePumpFailureAction(PumpAction):
    """Action class that will cause a pump failure (on/off -> failure, failure -> off)."""

    @staticmethod
    def new(target: int) -> "TogglePumpFailureAction":
        """Factory function for a `TogglePumpFailureAction`.

        Args:
            target (int | Literal["ab", "ba", "ca", "ec", "ea", "db", "fd", "fb"]): target pump

        Returns:
            TogglePumpFailureAction: the action.
        """
        return TogglePumpFailureAction(target=target)

    def __execute__(self, xml_state: XMLState):  # noqa
        pump_id = f"pump-{self.target}-button"
        # 0 -> 2, 1 -> 2, 2 -> 0
        new_state = "2 * (1 - {data-state} // 2)"
        xml_state.update(
            update(
                xpath=f"//*[@id='{pump_id}']",
                attrs={
                    "data-state": Expr(new_state),
                    # GOTCHA! data-state will be updated first (from above) and used to update fill! the order matters here.
                    "fill": Expr("{data-colors}[{data-state}]"),
                },
            )
        )


class PumpFuelAction(PumpAction):
    """Action class that will pump fuel from one tank to another."""

    flow: float
    XPATH_PUMP: ClassVar[str] = "//svg:rect[@id='pump-%s-button']"

    def __execute__(self, xml_state: XMLState):  # noqa
        if self.is_pump_on(xml_state, self.target):
            id_from = self.target[0]
            id_to = self.target[1]

            _from = _get_tank_data(xml_state, id_from)
            if _from["data-level"] <= 0:
                return  # there is no fuel to transfer
            _to = _get_tank_data(xml_state, id_to)
            remain = _to["data-capacity"] - _to["data-level"]
            if remain <= 0:
                return  # the tank is full

            # compute flow value and new levels
            # TODO potential weird bug here if _from["data-level"] is lower than flow for an infinite tank
            flow = min(_from["data-level"], remain, self.flow)
            new_to_level = _to["data-level"] + flow
            _update_tank_level(xml_state, id_to, _to, new_to_level)
            if id_from not in TANK_INF_IDS:
                # "from" tank is a normal tank, fuel should be removed
                new_from_level = _from["data-level"] - flow
                _update_tank_level(xml_state, id_from, _from, new_from_level)

    def is_pump_on(self, xml_state: XMLState, target: str) -> bool:
        """Is the given pump current in the "on" state?

        Args:
            xml_state (XMLState): environment state.
            target (str): target pump.

        Returns:
            bool: whether the pump is on.
        """
        xpath = PumpFuelAction.XPATH_PUMP % target
        data_state = xml_state.select(select(xpath=xpath, attrs=["data-state"]))[0][
            "data-state"
        ]
        return data_state == PumpAction.ON


class BurnFuelAction(XMLUpdateQuery):
    """Action class that will burn fuel in one of the main tanks (tank id = "a", "b" or "*" which indicates both tanks)."""

    target: str
    burn: float

    @field_validator("target", mode="before")
    @classmethod
    def _validate_target(cls, value: int | str) -> str:
        if isinstance(value, int):
            value = TANK_MAIN_IDS[value]
        if isinstance(value, str):
            if value in TANK_MAIN_IDS + [ALL]:
                return value
        raise ValueError(
            f"Invalid tank {value}, must be one of {TANK_MAIN_IDS + [ALL]}"
        )

    def __execute__(self, xml_state: XMLState):  # noqa
        targets = self._get_targets()
        for target in targets:
            values = _get_tank_data(xml_state, target)
            if values["data-level"] == 0:
                continue  # fuel is already 0, no fuel to burn
            new_level = max(values["data-level"] - self.burn, 0)
            _update_tank_level(xml_state, target, values, new_level)

    def _get_targets(self) -> list[str]:
        if self.target != ALL:
            targets = [self.target]
        else:
            targets = TANK_MAIN_IDS
        return targets


def _update_tank_level(
    xml_state: XMLState, tank: str, tank_data: dict[str, Any], new_level: float
):
    # updates the level of a tank
    new_height = tank_data["height"] * (new_level / tank_data["data-capacity"])
    new_y = tank_data["height"] - new_height
    xml_state.update(
        update(
            xpath=f"//*[@id='tank-{tank}-fuel']",
            attrs={"y": new_y, "height": new_height},
        )
    )
    xml_state.update(
        update(
            xpath=f"//*[@id='tank-{tank}']",
            attrs={"data-level": new_level},
        )
    )


def _get_tank_data(xml_state: XMLState, tank: str) -> dict[str, Any]:
    # getter for tank data
    return xml_state.select(
        select(
            xpath=f"//*[@id='tank-{tank}']",
            attrs=["data-level", "data-capacity", "height"],
        )
    )[0]



"""Module that implements the "tracking" task.

This files contains:
    - avatar actuator: `AvatarTrackingActuator`
    - agent actuator: `TrackingActuator`
    - actions: [`TargetMoveAction`]
"""

import math
import random
import time
from typing import Any
from pydantic import field_validator

from star_ray_xml import XMLState, Expr, update, select
from icua.event import KeyEvent, XMLUpdateQuery
from icua.utils import LOGGER
from icua.agent import Actuator, attempt

from ...utils._const import DEFAULT_KEY_BINDING  # TODO support other key bindings?

__all__ = ("AvatarTrackingActuator", "TrackingActuator", "TargetMoveAction")

DIRECTION_MAP = {
    "up": (0, -1),
    "down": (0, 1),
    "left": (-1, 0),
    "right": (1, 0),
}


class AvatarTrackingActuator(Actuator):
    """Actuator class that should be added to the Avatar when the tracking task is enabled. It allows the user to control the "target" in this task."""

    def __init__(
        self, target_speed: float = 5.0, *args: list[Any], **kwargs: dict[str, Any]
    ):
        """Constructor.

        Args:
            target_speed (float): the speed of the target (svg units per second).
            args (list[Any], optional): additional optional arguments.
            kwargs (dict[Any], optional): additional optional keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self._keys_pressed = set()
        self._prev_time = time.time()
        self._target_speed = target_speed

    def __attempt__(self) -> list["TargetMoveAction"]:
        """Attempt method that will attempt a `TargetMoveAction` to move the target according to the users input (if it has been provided since the last call to this method).

        Returns:
            list[TargetMoveAction]: the action
        """
        current_time = time.time()
        # this will contain the user action (KeyEvent)
        actions = []
        if len(self._keys_pressed) > 0:
            # compute speed based on time that has passed
            dt = current_time - self._prev_time
            speed = self._target_speed * dt
            # this will be normalised when the action is executed
            result = [0, 0]
            # compute the movement action based on the currently pressed keys
            for key in self._keys_pressed:
                direction = DIRECTION_MAP[DEFAULT_KEY_BINDING[key]]
                result[0] += direction[0]
                result[1] += direction[1]
            if result[0] != 0 or result[1] != 0:
                actions.append(TargetMoveAction(direction=tuple(result), speed=speed))
        self._prev_time = current_time
        return actions

    @attempt
    def attempt_key_event(self, user_action: KeyEvent) -> list:
        """Attempt method that takes a `KeyEvent` as input from the user. It is up to the avatar to provide this event to this actuator. The information is essential if the user is to control this task.

        Args:
            user_action (KeyEvent): the users keyboard action

        Returns:
            list: an empty list (no action is taken by this attempt method, see `AvatarTrackingActuator.__attempt__`.
        """
        if user_action.key.lower() in DEFAULT_KEY_BINDING:
            if user_action.status == KeyEvent.UP:
                self._keys_pressed.remove(user_action.key)
            elif user_action.status in (KeyEvent.DOWN, KeyEvent.HOLD):
                self._keys_pressed.add(user_action.key)
        return []

    # @attempt
    # def attempt_joystick_event(self, user_action: JoyStickEvent):
    #     # TODO If a joystick device is used (and supported elsewhere), this is where we would handle the action.
    #     # the handling should look similar to key events above but might require some work (if the input is continuous for example)
    #     raise NotImplementedError()


class TrackingActuator(Actuator):
    """Actuator class that will be part of a `ScheduledAgent` or any other agent that controls the evolution of the tracking task."""

    @attempt
    def move_target(
        self, direction: tuple[float, float] | int | float, speed: float
    ) -> "TargetMoveAction":
        """Move the tracking target in a given direction at a given speed.

        Args:
            direction (tuple[float, float] | int | float): direction to move.
            speed (float): speed to move (svg units per call to this function) # TODO perhaps this should be per second!

        Returns:
            TargetMoveAction: the action to move the tracking target.
        """
        # an angle was provided (in degrees), convert it to a direction vector
        if isinstance(direction, int | float):
            angle = math.radians(direction)
            direction = (math.sin(angle), math.cos(angle))
        return TargetMoveAction(direction=direction, speed=speed)

    @attempt
    def perturb_target(self, speed: float) -> "TargetMoveAction":
        """Move the tracking target in a random direction at a given speed.

        Args:
            speed (float): speed to move (svg units per call to this function) # TODO perhaps this should be per second!

        Returns:
            TargetMoveAction: the action to move the tracking target.
        """
        angle = (random.random() * 2 - 1) * math.pi
        direction = (math.sin(angle), math.cos(angle))
        return TargetMoveAction(direction=direction, speed=speed)


class TargetMoveAction(XMLUpdateQuery):
    """Action class that will update the tracking target position."""

    direction: tuple[float, float]
    speed: float

    @field_validator("direction", mode="before")
    @classmethod
    def _validate_direction(cls, value):
        if isinstance(value, list):
            value = tuple(value)  # a list is ok, convert it to a tuple
        if isinstance(value, tuple):
            if len(value) == 2:
                # normalise the direction
                d = math.sqrt(value[0] ** 2 + value[1] ** 2)
                if d == 0:
                    return (0.0, 0.0)
                return (float(value[0]) / d, float(value[1]) / d)
        raise ValueError(f"Invalid direction {value}, must be Tuple[float,float].")

    @field_validator("speed", mode="before")
    @classmethod
    def _validate_speed(cls, value):
        return float(value)

    def __execute__(self, state: XMLState):  # noqa
        if self.direction == (0.0, 0.0):
            LOGGER.warning(
                f"Attempted {TargetMoveAction.__name__} with direction (0,0)",
            )
            return
        dx = self.direction[0] * self.speed
        dy = self.direction[1] * self.speed
        # get properties of the tracking task
        properties = state.select(
            select(
                xpath="//svg:svg/svg:svg[@id='tracking']",
                attrs=["width", "height"],
            )
        )[0]
        # task bounds should limit the new position
        x1, y1 = (0.0, 0.0)
        x2, y2 = x1 + properties["width"], y1 + properties["height"]

        new_x = Expr("max(min({x} + {dx}, {x2} - {width}), {x1})", dx=dx, x1=x1, x2=x2)
        new_y = Expr("max(min({y} + {dy}, {y2} - {height}), {y1})", dy=dy, y1=x1, y2=y2)
        return state.update(
            update(
                xpath="//svg:svg/svg:svg/svg:svg[@id='tracking_target']",
                attrs=dict(x=new_x, y=new_y),
            )
        )



"""Module containing configuration classes."""

import json
from deepmerge import always_merger
from datetime import datetime
from pydantic import (
    BaseModel,
    Field,
    field_validator,
    # NonNegativeInt,
    PositiveInt,
    PositiveFloat,
    model_validator,
)
from typing import Any, ClassVar, Literal
from pathlib import Path
from icua.agent.actuator_guidance import ArrowGuidanceActuator, BoxGuidanceActuator
from star_ray.ui import WindowConfiguration
from icua.extras.eyetracking import EyetrackerBase, EyetrackerIOSensor
from ..utils import LOGGER


class GuidanceArrowConfiguration(BaseModel, validate_assignment=True):
    """Configuration relating to the arrow guidance."""

    enable: bool = Field(
        default=True,
        description="Whether to enable arrow guidance.",
    )
    mode: Literal["gaze", "mouse", "fixed"] = Field(
        default="gaze",
        description="The mode to use for the arrow guidance, `gaze` will use the current gaze position, `mouse` will use the current mouse position, `fixed` will use the fixed offset specified in `arrow_offset`.",
    )
    scale: float = Field(default=1.0, description="The scale of the arrow.")
    fill_color: str = Field(default="none", description="The fill colour of the arrow.")
    stroke_color: str = Field(
        default="#ff0000", description="The line colour of the arrow outline."
    )
    stroke_width: float = Field(
        default=4.0, description="The line width of the arrow outline."
    )
    offset: tuple[float, float] = Field(
        default=(80, 80),
        description='The offset of the arrow from its set position, this is the position of the arrow if in "fixed" mode.',
    )

    def validate_from_context(self, context: "Configuration"):  # noqa
        if not context.eyetracking.enable and self.mode == "gaze":
            raise ValueError(
                '`guidance.arrow.mode`: "gaze" is not supported when eyetracking is disabled.'
            )

    def to_actuator(self) -> ArrowGuidanceActuator:
        """Factory method for a guidance arrow actuator."""
        return ArrowGuidanceActuator(
            arrow_mode=self.mode,
            arrow_scale=self.scale,
            arrow_fill_color=self.fill_color,
            arrow_stroke_color=self.stroke_color,
            arrow_stroke_width=self.stroke_width,
            arrow_offset=self.offset,
        )


class GuidanceBoxConfiguration(BaseModel, validate_assignment=True):
    """Configuration relating to the guidance box."""

    enable: bool = Field(
        default=True, description="Whether to enable the guidance box."
    )
    stroke_color: str = Field(
        default="#ff0000", description="The line colour of the box outline."
    )
    stroke_width: float = Field(
        default=4.0, description="The line width of the box outline."
    )

    def validate_from_context(self, context: "Configuration"):  # noqa
        pass

    def to_actuator(self) -> BoxGuidanceActuator:
        """Factory method for a guidance box actuator."""
        return BoxGuidanceActuator(
            box_stroke_color=self.stroke_color,
            box_stroke_width=self.stroke_width,
        )


_TYPE_BREAK_TIES = Literal["random", "longest"]
_TYPE_GRACE_MODE = Literal["guidance_task", "guidance_any", "failure", "attention"]
_TYPE_ATTENTION_MODE = Literal["fixation", "gaze", "mouse"]


class GuidanceConfiguration(BaseModel, validate_assignment=True):
    """Configuration relating to guidance that may be provided to a user."""

    enable: bool = Field(
        default=True,
        description="Whether to enable guidance, if this is False then no guidance agent will be created. Only set this False if you do not plan to use AT ALL in your experiments, otherwise set `counter_factual` to True.",
    )
    counter_factual: bool = Field(
        default=False,
        description="Whether to show guidance to the user, if False then guidance agent will be configured NOT to display guidance but will still take actions for logging purposes (if they support this).",
    )
    break_ties: _TYPE_BREAK_TIES = Field(
        default="random",
        description="How to break ties if guidance may be shown on multiple tasks simultaneously.",
    )
    grace_mode: _TYPE_GRACE_MODE = Field(
        default="attention",
        description="Condition for which the `grace_period` is waiting. Options: `'guidance_task'` - time from last guidance a task, `'guidance_any'` - time from last guidance on _any_ task, `'failure'` - time from last failure a task, `'attention'` - time from when the user last attended to a task (according to `attention_mode`).",
    )
    attention_mode: _TYPE_ATTENTION_MODE = Field(
        default="fixation",
        description="Method used to track the attention of the user. Options: `'fixation'` - use eyetracking fixations, `'gaze'` - use eyetracking (fixation & saccade), `'mouse'` - use the mouse position.",
    )
    grace_period: float = Field(
        default=3.0,
        description="The grace period to use (seconds) - how long to wait before guidance is shown to the user, see also `grace_mode`.",
    )
    arrow: GuidanceArrowConfiguration = Field(
        default_factory=GuidanceArrowConfiguration,
        description="Configuration for displaying arrow guidance.",
    )
    box: GuidanceBoxConfiguration = Field(
        default_factory=GuidanceBoxConfiguration,
        description="Configuration for displaying box guidance.",
    )

    @field_validator("break_ties", mode="before")
    @classmethod
    def _validate_break_ties(cls, value: _TYPE_BREAK_TIES):
        if value.lower() not in ["random", "longest"]:
            raise ValueError(
                f"`guidance.break_ties` must be one of: {_TYPE_BREAK_TIES.__args__}"
            )
        return value.lower()

    @field_validator("attention_mode", mode="before")
    @classmethod
    def _validate_attention_mode(cls, value: _TYPE_ATTENTION_MODE):
        if value.lower() not in _TYPE_ATTENTION_MODE.__args__:
            raise ValueError(
                f"`guidance.attention_mode` must be one of: {_TYPE_ATTENTION_MODE.__args__}"
            )
        return value.lower()

    @field_validator("grace_mode", mode="before")
    @classmethod
    def _validate_grace_mode(cls, value: _TYPE_GRACE_MODE):
        if value.lower() not in _TYPE_GRACE_MODE.__args__:
            raise ValueError(
                f"`guidance.grace_mode` must be one of: {_TYPE_GRACE_MODE.__args__}"
            )
        return value.lower()

    @model_validator(mode="before")
    @classmethod
    def _validate(cls, data: dict[str, Any]):
        if data.get("counter_factual", False) and not data.get("enable", True):
            raise ValueError(
                "`guidance.counter_factual` cannot be True when `guidance.enable` is False. `guidance.enabled = False` should only be used when guidance is not going to be AT ALL in your experiments or analysis as it will completely disable this feature, otherwise, use `guidance.counter_factual = True`."
            )
        return data

    def validate_from_context(self, context: "Configuration"):  # noqa
        if self.enable:
            self.box.validate_from_context(context)
            self.arrow.validate_from_context(context)
            if (
                self.attention_mode in ["gaze", "fixation"]
                and not context.eyetracking.enable
            ):
                raise ValueError(
                    f'`guidance.attention_mode`: "{self.attention_mode}" is not supported when eyetracking is disabled.'
                )


class ExperimentConfiguration(BaseModel, validate_assignment=True):
    """Configuration relating to the experiment to be run."""

    id: str | None = Field(
        default=None, description="The unique ID of this experiment."
    )
    path: str = Field(
        # default_factory=lambda: Path("./").resolve().as_posix(),
        default="./",
        description="The full path to the directory containing task configuration files. If this is a relative path it is relative to the current working directory.",
    )
    duration: int = Field(
        default=-1,
        description="The duration (in seconds) of this experiment (the simulation will close after this time), a negative value will leave the simulation running forever.",
    )
    enable_video_recording: bool = Field(
        default=False,
        description="Whether to begin a screen recording of the experiment when the simulation starts, the video will be saved to the logging path when the experiment ends.",
    )
    enable_tasks: list[str] = Field(
        default=["system_monitoring", "resource_management", "tracking"],
        description="Which tasks to enable at the start of the simulation.",
    )
    meta: dict = Field(
        default={},
        description="Any additional meta data you wish to associate with this experiment.",
    )

    @field_validator("id", mode="before")
    @classmethod
    def _validate_id(cls, value: str | None):
        if value is None:
            LOGGER.warning("Configuration option: `experiment.id` was set to None")
        return value

    @field_validator("path", mode="before")
    @classmethod
    def _validate_path(cls, value: str):
        # we don't want to set it here, it should remain relative, just check that it exists!
        experiment_path = Path(value).expanduser().resolve()
        if not experiment_path.exists():
            raise ValueError(
                f"Configuration option `experiment.path` is not valid: `{experiment_path.as_posix()}` does not exist."
            )
        return value

    def validate_from_context(self, context: "Configuration"):  # noqa
        pass


class ParticipantConfiguration(BaseModel, validate_assignment=True):
    """Configuration relating to the participant (or user)."""

    id: str | None = Field(
        default=None,
        description="The unique ID of the participant that is taking part in the experiment.",
    )
    meta: dict = Field(
        default={},
        description="Any additional meta data you wish to associate with the participant.",
    )

    def validate_from_context(self, context: "Configuration"):  # noqa
        pass


class EyetrackingConfiguration(BaseModel, validate_assignment=True):
    """Configuration relating to eyetracking."""

    SUPPORTED_SDKS: ClassVar[tuple[str]] = ("tobii",)

    uri: str | None = Field(
        default=None,
        description="The eye tracker address (example: `'tet-tcp://172.28.195.1'`). If left unspecified `matbii` will attempt to find an eye tracker. For details on setting up eye tracking, consult the [wiki](https://github.com/dicelab-rhul/matbii/wiki/Eyetracking).",
    )
    sdk: str = Field(
        default="tobii",
        description="The eye tracking SDK to use, current options are: `['tobii']`.",
    )
    enable: bool = Field(default=False, description="Whether eye tracking is enabled.")
    moving_average_n: PositiveInt = Field(
        default=5,
        description="The window size to used to smooth eye tracking coordinates.",
    )
    velocity_threshold: PositiveFloat = Field(
        default=0.5,
        description="The threshold on gaze velocity which will determine saccades/fixations. This is defined in screen space, where the screen coordinates are normalised in the range [0,1]. **IMPORTANT NOTE:** different monitor sizes may require different values, unfortunately this is difficult to standardise without access to data on the gaze angle (which would be monitor size independent).",
    )

    def validate_from_context(self, context: "Configuration"):  # noqa
        pass

    @model_validator(mode="before")
    @classmethod
    def _validate_model(cls, data: dict[str, Any]):
        if "enabled" in data:
            LOGGER.warning(
                "Configuration option: `eyetracking.enabled` is deprecated and will be removed in the future, please use `eyetracking.enable` instead."
            )
            data["enable"] = data["enabled"]
            del data["enabled"]
        return data

    @field_validator("sdk", mode="before")
    @classmethod
    def _validate_sdk(cls, value: str):
        if value not in EyetrackingConfiguration.SUPPORTED_SDKS:
            raise ValueError(
                f"Eyetracker SDK: {value} is not supported, must be one of {EyetrackingConfiguration.SUPPORTED_SDKS}"
            )
        return value

    def new_eyetracking_sensor(self) -> EyetrackerIOSensor | None:
        """Factory method for an eyetracking sensor.

        Returns:
            EyetrackerIOSensor | None: the sensor, created based on this eyetracking configuration.
        """
        if self.enable:
            eyetracker = self.new_eyetracker()
            return EyetrackerIOSensor(
                eyetracker, self.velocity_threshold, self.moving_average_n
            )
        return None

    def new_eyetracker(self) -> EyetrackerBase | None:
        """Factory method for an eyetracker.

        Returns:
            EyetrackerBase | None: the eyetracker created based on this eyetracking configuration.
        """
        if self.enable:
            if self.sdk == "tobii":
                return self._new_tobii_eyetracker()
            else:
                raise ValueError(
                    f"Eyetracker SDK: {self.sdk} is not supported, must be one of {EyetrackingConfiguration.SUPPORTED_SDKS}"
                )
        return None

    def _new_tobii_eyetracker(self) -> EyetrackerBase:
        from icua.extras.eyetracking import tobii  # this may fail!

        return tobii.TobiiEyetracker(uri=self.uri)


class LoggingConfiguration(BaseModel, validate_assignment=True):
    """Configuration relating to logging (including event logging)."""

    level: str = Field(
        default="INFO",
        description="The logging level to use: ['DEBUG', 'INFO', 'WARNING', 'ERROR'], this will not affect event logging.",
    )
    path: str = Field(
        default="./logs/",
        description="The path to the directory where log files will be written.",
    )

    @field_validator("level", mode="before")
    @classmethod
    def _validate_level(cls, value: str):
        _value = value.upper()
        if _value not in ["DEBUG", "INFO", "WARNING", "ERROR"]:
            raise ValueError(
                f"Configuration option `logging.level` is invalid: `{value}` must be one of: ['DEBUG', 'INFO', 'WARNING', 'ERROR']"
            )
        return _value


class UIConfiguration(BaseModel, validate_assignment=True):
    """Configuration relating to rendering and the UI."""

    width: PositiveInt = Field(
        default=810,
        description="The width of the canvas used to render the tasks.",
    )
    height: PositiveInt = Field(
        default=680,
        description="The height of the canvas used to render the tasks.",
    )
    # size: tuple[PositiveInt, PositiveInt] = Field(
    #     default=(810, 680),
    #     description="The width and height of the canvas used to render the tasks. This should fully encapsulate all task elements. If a task appears to be off screen, try increasing this value.",
    # )
    # offset: tuple[NonNegativeInt, NonNegativeInt] = Field(
    #     default=(0, 0),
    #     description="The x and y offset used when rendering the root UI element, can be used to pad the top/left of the window.",
    # )

    @model_validator(mode="before")
    @classmethod
    def _validate_model(cls, data: dict[str, Any]):
        if "size" in data:
            LOGGER.warning(
                "Configuration option: `ui.size` is deprecated and will be removed in the future, please use `ui.width` and `ui.height` instead."
            )
            data["width"] = data["size"][0]
            data["height"] = data["size"][1]
            del data["size"]
        if "offset" in data:
            LOGGER.warning(
                "Configuration option: `ui.offset` is deprecated and has been removed as an option, please instead modify the coordinates of each task to reflect the desired offset."
            )
            del data["offset"]
        return data


def _default_window_configuration_factory():
    window_config = WindowConfiguration()
    window_config.width = 810
    window_config.height = 680
    window_config.title = "icua matbii"
    return window_config


class Configuration(BaseModel, validate_assignment=True):
    """Main configuration class for the default entry point."""

    experiment: ExperimentConfiguration = Field(default_factory=ExperimentConfiguration)
    participant: ParticipantConfiguration = Field(
        default_factory=ParticipantConfiguration
    )
    guidance: GuidanceConfiguration = Field(default_factory=GuidanceConfiguration)
    window: WindowConfiguration = Field(
        default_factory=_default_window_configuration_factory
    )
    eyetracking: EyetrackingConfiguration = Field(
        default_factory=EyetrackingConfiguration
    )
    logging: LoggingConfiguration = Field(default_factory=LoggingConfiguration)
    ui: UIConfiguration = Field(default_factory=UIConfiguration)

    @staticmethod
    def from_file(
        path: str | Path | None, context: dict[str, Any] | None = None
    ) -> "Configuration":
        """Factory that will build `Configuration` from a .json file.

        Args:
            path (str | Path | None): path to config file.
            context (dict[str, Any] | None, optional): additional context (to override file content). Defaults to None.

        Returns:
            Configuration: resulting configuration.
        """
        context = context if context else {}
        if path:
            path = Path(path).expanduser().resolve().as_posix()
            LOGGER.info(f"Using config file: {path}")
            with open(path) as f:
                data = json.load(f)
                data = always_merger.merge(data, context)
                result = Configuration.model_validate(data)
                result.validate_from_context()  # check consistency of nested fields
                return result
        else:
            LOGGER.info("No config file was specified, using default configuration.")
            return Configuration()

    @staticmethod
    def initialise_logging(config: "Configuration") -> "Configuration":
        """Initialises logging for the given run. This will set logging options and set the config logging path which should be used throughout `matbii` to log information that may be relevant for experiment post-analysis. The configuration passed here will also be logged to the `configuration.json` file in the logging path.

        The logging path will be derived: `<config.experiment.id>/<config.participant.id>` if these values are present, otherwise a timestamp will be used to make the logging path unique. If the two ids are given then they are assumed to be unique (they represent a single trial for a participant).

        Args:
            config (Configuration): configuration

        Raises:
            FileExistsError: if the derived logging path already exists.

        Returns:
            Configuration: the configuration (with updated path variables - modified in place)
        """
        # set logging level
        LOGGER.set_level(config.logging.level)

        # set the logger path
        path = Path(config.logging.path).expanduser().resolve()

        # create the full path
        full_path = path
        if config.experiment.id:
            full_path = full_path / config.experiment.id
        if config.participant.id:
            full_path = full_path / config.participant.id

        if path == full_path:
            full_path = full_path / datetime.now().strftime("%Y%m%d%H%M%S%f")[:-3]

        if full_path.exists():
            raise FileExistsError(
                f"Logging path: {path} already exists, perhaps you have miss-specified `experiment.id` or `participant.id`?\n    See <TODO LINK> for details."
            )

        full_path.mkdir(parents=True)
        LOGGER.debug(f"Logging to: {full_path.as_posix()}")

        # log the configuration that is in use
        with open(full_path / "configuration.json", "w") as f:
            f.write(config.model_dump_json(indent=2))

        config.logging.path = full_path.as_posix()
        return config

    def validate_from_context(self):  # noqa
        self.guidance.validate_from_context(self)
        self.eyetracking.validate_from_context(self)



"""Package that defines the matbii resource management task."""

from cerberus import rules_set_registry as _rsr
from .resource_management import (
    ResourceManagementActuator,
    AvatarResourceManagementActuator,
    SetPumpAction,
    BurnFuelAction,
    PumpFuelAction,
    TogglePumpAction,
    TogglePumpFailureAction,
)

__all__ = (
    "ResourceManagementActuator",
    "AvatarResourceManagementActuator",
    "SetPumpAction",
    "TogglePumpAction",
    "BurnFuelAction",
    "PumpFuelAction",
    "TogglePumpFailureAction",
)

# define type for pump state in configuration files, the values an be "on", "off", "failure", see SetPumpAction.coerce_pump_state
_rsr.add(
    "pump_state",
    {
        "type": "integer",
        "allowed": [0, 1, 2],
        "default": "off",
        "coerce": SetPumpAction.coerce_pump_state,
    },
)



"""Module contains a guidance sensor for tracking the "tracking" task acceptability, see `TrackingTaskAcceptabilitySensor` documentation for details."""

from typing import Any
from star_ray.event.observation_event import Observation
from star_ray_xml import select, Select
from .sensor_guidance import TaskAcceptabilitySensor

from ..utils._const import TASK_ID_TRACKING, tracking_box_id, tracking_target_id


class TrackingTaskAcceptabilitySensor(TaskAcceptabilitySensor):
    """Guidance sensor for the tracking task."""

    _BOX_ID = tracking_box_id()
    _TARGET_ID = tracking_target_id()

    def __init__(self, *args: list[Any], **kwargs: dict[str, Any]):
        """Constructor.

        Args:
            args (list[Any], optional): additional optional arguments.
            kwargs (dict[Any], optional): additional optional keyword arguments.
        """
        super().__init__(TASK_ID_TRACKING, *args, **kwargs)
        # initialise beliefs about the tracking tasks
        self.beliefs[TrackingTaskAcceptabilitySensor._BOX_ID] = dict()
        self.beliefs[TrackingTaskAcceptabilitySensor._TARGET_ID] = dict()

    def is_acceptable(self, task: str = None, **kwargs: dict[str, Any]) -> bool:  # noqa
        if task is None or task == self.task_name:
            return self.is_tracking_acceptable()
        else:
            raise KeyError(
                f"Invalid subtask: {task}, doesn't exist for task {self.task_name}"
            )

    def is_tracking_acceptable(self) -> bool:
        """Determines whether the tracking task is in an acceptable state.

        Acceptable: "target" is within the central box of the task.
        Unacceptable: otherwise.

        Raises:
            ValueError: if there is missing observational data about the task - this may mean the task is not active.

        Returns:
            bool: whether the task is in an acceptable state.
        """
        try:
            target = self.beliefs.get(TrackingTaskAcceptabilitySensor._TARGET_ID, None)
        except KeyError:
            raise KeyError(
                f"Failed to determine acceptability of task: '{TASK_ID_TRACKING}'.\n-- Missing beliefs for elements: '{TrackingTaskAcceptabilitySensor._TARGET_ID}'"
            )
        try:
            box = self.beliefs.get(TrackingTaskAcceptabilitySensor._BOX_ID, None)
        except KeyError:
            raise KeyError(
                f"Failed to determine acceptability of task: '{TASK_ID_TRACKING}'.\n-- Missing beliefs for elements: '{TrackingTaskAcceptabilitySensor._BOX_ID}'"
            )
        return TrackingTaskAcceptabilitySensor.is_point_in_rectangle(
            target["xy"], box["tl"], box["br"]
        )

    def on_observation(self, observation: Observation):
        """Update beliefs about the tracking task based on the incoming observation.

        Args:
            observation (Observation): observation
        """
        for data in observation.values:
            try:
                if data["id"] == TrackingTaskAcceptabilitySensor._TARGET_ID:
                    tl = (data["x"], data["y"])
                    size = (data["width"], data["height"])
                    xy = (tl[0] + size[0] / 2, tl[1] + size[1] / 2)
                    self.beliefs[TrackingTaskAcceptabilitySensor._TARGET_ID]["xy"] = xy
                    self.beliefs[TrackingTaskAcceptabilitySensor._TARGET_ID]["size"] = (
                        size
                    )
                elif data["id"] == TrackingTaskAcceptabilitySensor._BOX_ID:
                    tl = (data["x"], data["y"])
                    self.beliefs[TrackingTaskAcceptabilitySensor._BOX_ID]["tl"] = tl
                    self.beliefs[TrackingTaskAcceptabilitySensor._BOX_ID]["br"] = (
                        tl[0] + data["width"],
                        tl[1] + data["height"],
                    )
            except KeyError:
                raise KeyError(
                    f"Observation: {observation} doesn't contain the required `id` attribute."
                )

    def sense(self) -> list[Select]:
        """Generates the sense actions that are required for checking whether the tracking task is in an acceptable state.

        The actions will request the following data:
        - the bounds of the target element.
        - the bounds of the central box of the tracking task.

        Returns:
            list[Select]: list of sense actions to take.
        """
        return [
            select(
                xpath=f"//*[@id='{tracking_target_id()}']",
                attrs=["id", "x", "y", "width", "height"],
            ),
            select(
                xpath=f"//*[@id='{tracking_box_id()}']",
                attrs=["id", "x", "y", "width", "height"],
            ),
        ]

    @staticmethod
    def is_point_in_rectangle(
        point: tuple[float, float],
        rect_min: tuple[float, float],
        rect_max: tuple[float, float],
    ) -> bool:
        """Checks wether the given `point` is within the rectangle as specified by the top left and bottom right coordinate.

        Args:
            point (tuple[float, float]): point to check.
            rect_min (tuple[float,float]): top left.
            rect_max (tuple[float,float]): bottom right.

        Returns:
            bool: whether the point is in the rectangle.
        """
        px, py = point
        min_x, min_y = rect_min
        max_x, max_y = rect_max
        return min_x <= px <= max_x and min_y <= py <= max_y



"""Module contains a guidance sensor for tracking the "resource management" task acceptability, see `ResourceManagementTaskAcceptabilitySensor` documentation for details."""

from typing import Any
from functools import partial
from star_ray_xml import select, Select
from .sensor_guidance import TaskAcceptabilitySensor
from ..utils._const import TASK_ID_RESOURCE_MANAGEMENT, tank_id, tank_level_id


class ResourceManagementTaskAcceptabilitySensor(TaskAcceptabilitySensor):
    """Guidance sensor for the resource management task.

    This sensor tracks a number of sub-tasks:

    - "system_monitoring.tank-a"
    - "system_monitoring.tank-b"

    The acceptability of these sub-tasks can be checked by calling the method: `ResourceManagementTaskAcceptabilitySensor.is_tank_acceptable`.

    Otherwise follow the `TaskAcceptabilitySensor` API.
    """

    def __init__(self, *args: list[Any], **kwargs: dict[str, Any]):
        """Constructor.

        Args:
            args (list[Any], optional): additional optional arguments.
            kwargs (dict[Any], optional): additional optional keyword arguments.
        """
        super().__init__(TASK_ID_RESOURCE_MANAGEMENT, *args, **kwargs)
        self._is_subtask_acceptable_map = {
            f"{self.task_name}.tank-a": partial(self.is_tank_acceptable, "a"),
            f"{self.task_name}.tank-b": partial(self.is_tank_acceptable, "b"),
        }

    def is_acceptable(self, task: str = None, **kwargs: dict[str, Any]) -> bool:  # noqa
        if task is None or task == self.task_name:
            return all([x() for x in self._is_subtask_acceptable_map.values()])
        else:
            is_acceptable = self._is_subtask_acceptable_map.get(task, None)
            if is_acceptable is None:
                raise KeyError(
                    f"Invalid subtask: {task}, doesn't exist for task {self.task_name}"
                )
            else:
                return is_acceptable()

    def is_tank_acceptable(self, _id: str) -> bool:
        """Whether the given tank is in an acceptable state.

        Acceptable: the fuel level lies in the required range.
        Unacceptable: otherwise.

        Args:
            _id (int): the id of the tank ("a" or "b")

        Returns:
            bool: True if the tank is in an acceptable state, False otherwise.
        """
        tank = self.beliefs[tank_id(_id)]
        tank_level = self.beliefs[tank_level_id(_id)]

        fuel_capacity = tank["data-capacity"]
        fuel_level = tank["data-level"]
        acceptable_level = tank_level["data-level"] * fuel_capacity
        acceptable_range2 = (tank_level["data-range"] * fuel_capacity) / 2
        return (
            fuel_level >= acceptable_level - acceptable_range2
            and fuel_level <= acceptable_level + acceptable_range2
        )

    def sense(self, tank_ids: tuple[str, ...] = ("a", "b")) -> list[Select]:
        """Generates the sense actions that are required for checking whether the resource management task is in an acceptable state.

        The actions will request the following data:
        - the current fuel level and capacity of each main tank ("a" and "b")
        - the acceptable range of fuel for each main tank ("a" and "b")

        Args:
            tank_ids (tuple[str, str], optional): the ids of the tanks to check, defaults to ("a", "b").

        Returns:
            list[Select]: list of sense actions to take.
        """
        # interested in the fuel levels of the main tanks in the Resource Management Task
        tanks = [tank_id(i) for i in tank_ids]
        tank_levels = [tank_level_id(i) for i in tank_ids]
        tank_selects = [
            select(
                xpath=f"//*[@id='{id}']", attrs=["id", "data-capacity", "data-level"]
            )
            for id in tanks
        ]
        tank_level_selects = [
            select(xpath=f"//*[@id='{id}']", attrs=["id", "data-level", "data-range"])
            for id in tank_levels
        ]
        return [*tank_selects, *tank_level_selects]



"""Functions for extracting task events from an event log file."""

import warnings
import pandas as pd
import numpy as np
from functools import partial
from star_ray.agent.component.component import Component
from star_ray_xml import XMLState, Insert, Update, Replace, Delete
from star_ray_pygame import SVGAmbient
from icua.event import (
    Event,
    RenderEvent,
    MouseButtonEvent,
    KeyEvent,
    Select,
    UserInputEvent,
)
from icua.extras.analysis import EventLogParser

# these sensors are going to be used to get the relevant state information via their sense actions
from ...guidance import (
    SystemMonitoringTaskAcceptabilitySensor,
    TrackingTaskAcceptabilitySensor,
    # ResourceManagementTaskAcceptabilitySensor, # the sense actions are defined here...
)

# used to create resource management sense actions
from ...utils._const import (
    tank_ids,
    pump_ids,
)

# system monitoring events
from ...tasks import (
    SetSliderAction,
    SetLightAction,
    ToggleLightAction,
    TargetMoveAction,
    SetPumpAction,
    BurnFuelAction,
    PumpFuelAction,
    TogglePumpAction,
    TogglePumpFailureAction,
)


# def get_resource_management_task_events(
#     parser: EventLogParser,
#     events: list[tuple[float, Event]],
#     norm: float | int = np.inf,
# ) -> pd.DataFrame:
#     """Extracts useful data for the resource management task from the event log.

#     Columns:
#         - timestamp: float - the (logging) timestamp of the event
#         - frame: int - the frame number of the event, events with a frame number of 0 happen BEFORE the first frame is rendered to the user.
#         - user: bool - whether the event was triggered by the user or not.
#         - tank-a: float - tank a level
#         - tank-b: float - tank b level
#         - tank-c: float - tank c level
#         - tank-d: float - tank d level
#         - tank-e: float - tank e level
#         - tank-f: float - tank f level

#     Args:
#         parser (EventLogParser): parser used to parse the event log file.
#         events (list[tuple[float, Event]]): list of events that were parsed from the event log file.
#         norm (float | int, optional): the norm to use for the distance metric, either "inf" for the max norm or an integer for the p-norm.

#     Returns:
#         pd.DataFrame: dataframe with columns: ["timestamp", "frame", "user", *"tanks-{i}", *"pumps-{ij}"]
#     """
#     from matbii.utils import LOGGER

#     LOGGER.warning(
#         "`get_resource_management_task_events` is not implemented yet, the result will be an empty dataframe."
#     )
#     df = None
#     if df is None:
#         columns = [
#             "timestamp",
#             "frame",
#             "user",
#             "x",
#             "y",
#             "distance",
#         ]
#         return pd.DataFrame(columns=columns)
#     return df


def get_resource_management_task_events(
    parser: EventLogParser,
    events: list[tuple[float, Event]],
    norm: float | int = np.inf,
) -> pd.DataFrame:
    """Extracts useful data for the resource management task from the event log.

    Columns:
        - timestamp: float - the (logging) timestamp of the event
        - frame: int - the frame number of the event, events with a frame number of 0 happen BEFORE the first frame is rendered to the user.
        - user: bool - whether the event was triggered by the user or not.
        - tank-a: float - tank a level
        - tank-b: float - tank b level
        - tank-c: float - tank c level
        - tank-d: float - tank d level
        - tank-e: float - tank e level
        - tank-f: float - tank f level

    Args:
        parser (EventLogParser): parser used to parse the event log file.
        events (list[tuple[float, Event]]): list of events that were parsed from the event log file.
        norm (float | int, optional): the norm to use for the distance metric, either "inf" for the max norm or an integer for the p-norm.

    Returns:
        pd.DataFrame: dataframe with columns: ["timestamp", "frame", "user", *"tanks-{i}", *"pumps-{ij}"]
    """
    fevents = parser.filter_events(
        events,
        (
            Insert,
            Delete,
            Update,
            Replace,
            SetPumpAction,
            BurnFuelAction,
            PumpFuelAction,
            TogglePumpAction,
            TogglePumpFailureAction,
            KeyEvent,
            MouseButtonEvent,
            RenderEvent,
        ),
    )

    # # sort the events by their log timestamp
    fevents = parser.sort_by_timestamp(fevents)

    def _sense(state: XMLState):
        # sense_actions = ResourceManagementTaskAcceptabilitySensor().sense()
        # sense actions for tank states
        sense_actions = [
            Select(xpath=f"//*[@id='{id}']", attrs=["id", "data-level"])
            for id in tank_ids()
        ]
        result = sense(state, sense_actions)
        if result is None:
            result = dict()
        result_tanks = {k: v["data-level"] for k, v in result.items()}

        # actions for pump states
        sense_actions = [
            Select(xpath=f"//*[@id='{id}-button']", attrs=["id", "data-state"])
            for id in pump_ids()
        ]
        result = sense(state, sense_actions)
        # print(result)
        if result is None:
            result = dict()
        result_pumps = {k.rsplit("-", 1)[0]: v["data-state"] for k, v in result.items()}
        if len(result_pumps) == 0:
            return None
        return {**result_tanks, **result_pumps}

    df = _get_task_dataframe(fevents, _sense)
    if df is None:
        columns = [
            "timestamp",
            "frame",
            "user",
            *tank_ids(),
            *pump_ids(),
        ]
        return pd.DataFrame(columns=columns)

    # TODO combine rows if the timestamp matches
    return df


def get_tracking_task_events(
    parser: EventLogParser,
    events: list[tuple[float, Event]],
    norm: float | int = np.inf,
) -> pd.DataFrame:
    """Extracts useful data for the tracking task from the event log.

    Columns:
        - timestamp: float - the (logging) timestamp of the event
        - frame: int - the frame number of the event, events with a frame number of 0 happen BEFORE the first frame is rendered to the user.
        - user: bool - whether the event was triggered by the user or not.
        - x: float - the x coordinate of the tracking target
        - y: float - the y coordinate of the tracking target
        - distance: float - the distance to the center of the task (according to the given `norm`).

    Args:
        parser (EventLogParser): parser used to parse the event log file.
        events (list[tuple[float, Event]]): list of events that were parsed from the event log file.
        norm (float | int, optional): the norm to use for the distance metric, either "inf" for the max norm or an integer for the p-norm.

    Returns:
        pd.DataFrame: dataframe with columns: ["timestamp", "frame", "user", "x", "y", "distance"]
    """
    fn_norm = partial(np.linalg.norm, ord=norm)
    # collect all the relevant events for this task
    fevents = parser.filter_events(
        events,
        (
            Insert,
            Delete,
            Update,
            Replace,
            TargetMoveAction,
            KeyEvent,
            MouseButtonEvent,
            RenderEvent,
        ),
    )

    # # sort the events by their log timestamp
    fevents = parser.sort_by_timestamp(fevents)

    def _sense(state: XMLState):
        sense_actions = TrackingTaskAcceptabilitySensor().sense()
        result = sense(state, sense_actions)
        if result is None:
            return None
        # TODO compute distance to the center of the task
        bw, bh = (
            result[TrackingTaskAcceptabilitySensor._BOX_ID]["width"],
            result[TrackingTaskAcceptabilitySensor._BOX_ID]["height"],
        )
        bx, by = (
            result[TrackingTaskAcceptabilitySensor._BOX_ID]["x"] + bw / 2,
            result[TrackingTaskAcceptabilitySensor._BOX_ID]["y"] + bh / 2,
        )
        tw, th = (
            result[TrackingTaskAcceptabilitySensor._TARGET_ID]["width"],
            result[TrackingTaskAcceptabilitySensor._TARGET_ID]["height"],
        )
        tx, ty = (
            result[TrackingTaskAcceptabilitySensor._TARGET_ID]["x"] + tw / 2,
            result[TrackingTaskAcceptabilitySensor._TARGET_ID]["y"] + th / 2,
        )
        return dict(
            x=tx,
            y=ty,
            distance=fn_norm((tx - bx, ty - by)),
        )

    df = _get_task_dataframe(fevents, _sense)
    if df is None:
        columns = [
            "timestamp",
            "frame",
            "user",
            "x",
            "y",
            "distance",
        ]
        return pd.DataFrame(columns=columns)
    return df


def get_system_monitoring_task_events(
    parser: EventLogParser,
    events: list[tuple[float, Event]],
) -> pd.DataFrame:
    """Extracts useful data for the system monitoring task from the event log.

    Columns:
        - timestamp: float - the (logging) timestamp of the event
        - frame: int - the frame number of the event, events with a frame number of 0 happen BEFORE the first frame is rendered to the user.
        - user: bool - whether the event was triggered by the user (True) or not (False).
        - light-1: int - the state of light-1
        - light-2: int - the state of light-2
        - slider-1: int - the state of slider-1
        - slider-2: int - the state of slider-2
        - slider-3: int - the state of slider-3
        - slider-4: int - the state of slider-4

    NOTE: the number of sliders/lights is assumed to be 4/2 respectively, any alternative settings will not work here - this may be fixed in future versions.

    Args:
        parser (EventLogParser): parser used to parse the event log file.
        events (list[tuple[float, Event]]): list of events that were parsed from the event log file.

    Returns:
        pd.DataFrame: dataframe with columns: ["timestamp", "frame", "user", "light-1", "light-2", "slider-1", "slider-2", "slider-3", "slider-4"]
    """
    # these are raw events that may be triggered internally by matbii
    # insert events insert the tasks into the xml state at the start of the simulation
    fevents = parser.filter_events(
        events,
        (
            Insert,
            Delete,
            Replace,
            Update,
            SetLightAction,
            ToggleLightAction,
            SetSliderAction,
            KeyEvent,
            MouseButtonEvent,
            RenderEvent,
        ),
    )
    # sort the events by their timestamp
    fevents = parser.sort_by_timestamp(fevents)

    def _rename_column(k: str):
        return "-".join(k.split("-")[:2])

    def _sense(state: XMLState):
        # sense actions to get relevant data from the state
        sense_actions = SystemMonitoringTaskAcceptabilitySensor().sense()
        result = sense(state, sense_actions)
        if result is None:
            return None
        return {
            _rename_column(k): v["data-state"]
            for k, v in result.items()
            if "data-state" in v
        }

    df = _get_task_dataframe(fevents, _sense)
    if df is None:
        columns = [
            "timestamp",
            "frame",
            "user",
            "light-1",
            "light-2",
            "slider-1",
            "slider-2",
            "slider-3",
            "slider-4",
        ]
        return pd.DataFrame(columns=columns)
    return df


def _get_task_dataframe(
    fevents, fn_sense, user_input_event_type: type = UserInputEvent
):
    """Used internally to build a dataframe for a task."""
    # use the default state, the actual size etc. of the svg is not important for our purposes.
    # we only want to track the task events (which do not depend on the svg or window config.)
    xml_state = SVGAmbient([]).get_state()
    # sort the events by their log timestamp
    fevents = EventLogParser.sort_by_timestamp(fevents)

    def _get_task_events(fevents: list[tuple[float, Event]], avatar_ids: set[int]):
        """Execute the events in order and yield the current task state."""
        frame = 0
        for i, (t, event) in enumerate(fevents):
            if isinstance(event, RenderEvent):
                frame += 1
                continue
            if isinstance(event, user_input_event_type):
                _, avatar_id = Component.unpack_source(event)
                avatar_ids.add(avatar_id)
                continue
            event.__execute__(xml_state)  # apply the event to the state
            result = fn_sense(xml_state)
            if result is None:
                # ignore this if the the task is not yet set up, its only a probably if you see it spammed lots!
                warnings.warn(
                    f"Error sensing data for event {i} of type {type(event)}."
                )
                continue
            if event.source is None:
                warnings.warn(
                    f"Event {i} of type {type(event)} has no source, your log file is out of date."
                )
                result["agent"] = 0
            else:
                _, result["agent"] = Component.unpack_source(event)
            result["timestamp"] = t
            result["frame"] = frame
            yield result

    avatar_ids = set()

    df = pd.DataFrame(_get_task_events(fevents, avatar_ids))
    if len(avatar_ids) > 1:
        raise ValueError(
            "Multiple avatar ids inferred, this should not happen, has there been a change to the event logging system?"
        )
    if len(df) == 0:
        return None

    df["user"] = df["agent"] == next(iter(avatar_ids), float("nan"))
    df.drop(columns=["agent"], inplace=True)
    # organise columns
    start_columns = ["timestamp", "frame", "user"]
    state_columns = sorted(list(set(df.columns) - set(start_columns)))
    return df[start_columns + state_columns]


def sense(state: XMLState, sense_actions: list[Select]):
    """Sense data from the state using the provided sense actions."""

    def _sense(state: XMLState, sense_actions: list[Select]):
        for action in sense_actions:
            yield action.__execute__(state)

    data = dict()
    try:
        for obs in _sense(state, sense_actions):
            for value in obs:
                id = value.pop("id")
                data[id] = value
    except Exception:
        return None
    return data



from pathlib import Path

# constants defining the task ids
TASK_ID_TRACKING = "tracking"
TASK_ID_SYSTEM_MONITORING = "system_monitoring"
TASK_ID_RESOURCE_MANAGEMENT = "resource_management"

# path of the tasks (in this package)
PACKAGE_PATH = Path(__file__).resolve().parent.parent  # root of the package
TASKS_PATH = PACKAGE_PATH / "tasks"

# tasks ids, add to this when/if more tasks are implemented
TASKS = [TASK_ID_TRACKING, TASK_ID_SYSTEM_MONITORING, TASK_ID_RESOURCE_MANAGEMENT]
# paths of each task (within this package), ensure that the task id matches the folder name in `matbii.tasks`
TASK_PATHS = {t: (TASKS_PATH / t).as_posix() for t in TASKS}
for task, path in TASK_PATHS.items():
    if not Path(path).exists():
        raise FileNotFoundError(f"Failed to find task: {task} at path: {path}")

# enable these tasks by default
DEFAULT_ENABLED_TASKS = [
    TASK_ID_TRACKING,
    TASK_ID_SYSTEM_MONITORING,
    TASK_ID_RESOURCE_MANAGEMENT,
]

DEFAULT_KEY_BINDING = {
    "arrowright": "right",
    "arrowleft": "left",
    "arrowup": "up",
    "arrowdown": "down",
    "right": "right",
    "left": "left",
    "up": "up",
    "down": "down",
}

ALTERNATE_KEY_BINDING = {
    "d": "right",
    "a": "left",
    "w": "up",
    "s": "down",
}


# PYGAME_CONST = dict(
#     DEFAULT_KEY_BINDING={
#         "right": "right",
#         "left": "left",
#         "up": "up",
#         "down": "down",
#     },
#     ALTERNATE_KEY_BINDING={
#         "d": "right",
#         "a": "left",
#         "w": "up",
#         "s": "down",
#     },
# )


###### constant IDs for tasks, these are defined in the task svgs, but it is useful to have access. ######


def tank_id(tank: str):
    assert tank in ("a", "b", "c", "d", "e", "f")
    return f"tank-{tank}"


def tank_level_id(tank: str):
    assert tank in ("a", "b", "c", "d", "e", "f")
    return f"tank-{tank}-level"


def tank_ids() -> tuple[str]:
    return (
        tank_id("a"),
        tank_id("b"),
        tank_id("c"),
        tank_id("d"),
        tank_id("e"),
        tank_id("f"),
    )


def pump_id(tank1: str, tank2: str):
    assert tank1 in ("a", "b", "c", "d", "e", "f")
    assert tank2 in ("a", "b", "c", "d", "e", "f")
    return f"pump-{tank1}{tank2}"


def pump_ids() -> tuple[str]:
    return (
        pump_id("a", "b"),
        pump_id("b", "a"),
        pump_id("f", "d"),
        pump_id("f", "b"),
        pump_id("d", "b"),
        pump_id("e", "c"),
        pump_id("e", "a"),
        pump_id("c", "a"),
    )


def slider_id(slider: int):
    assert slider in (1, 2, 3, 4)
    return f"slider-{slider}-button-container"


def slider_incs_id(slider: int):
    assert slider in (1, 2, 3, 4)
    return f"slider-{slider}-incs"


def light_id(light: int):
    assert light in (1, 2)
    return f"light-{light}-button"


def tracking_box_id():
    return "tracking_box"


def tracking_target_id():
    return "tracking_target"



"""This module contains various scripts (functions) that can be used with the argument --script. Typically they will produce some output file (e.g. a csv or plot) for some data of interest from a log file. The avaliable functions are intended to be called from the command line and not to to be used directly. The make use of the `matbii.extras.analysis` module to produce files and plots."""

import argparse
import numpy as np
import pandas as pd
from typing import Any, Literal
import matplotlib.pyplot as plt
from pathlib import Path
from ..utils import LOGGER
from ..config import Configuration


TRACKING_COLOR = "#4363d8"
RESOURCE_MANAGEMENT_COLOR = "#3cb44b"
SYSTEM_MONITORING_COLOR = "#e6194B"


def summary(**kwargs: dict[str, Any]) -> None:
    """Produce a summary from the given logging directory."""
    parser = argparse.ArgumentParser(
        description="Produce a summary from the given logging directory."
    )
    parser.add_argument(
        "--path", type=str, required=True, help="The path to the logging directory."
    )
    parser.add_argument(
        "--output",
        type=str,
        required=False,
        help="The path to the output directory, if left unspecified files will be written to <--path>/summary.",
    )
    args, _ = parser.parse_known_intermixed_args()  # ignore unknown args
    path = Path(args.path)
    log_file, config_file = _validate_logging_path(path)
    config = _load_config(config_file, context=kwargs)

    if args.output:
        output_dir = Path(args.output)
    else:
        output_dir = path / "summary"
    output_dir.mkdir(parents=True, exist_ok=True)
    _summary(log_file, config, output_dir)


# ============================================= #
# ================ INTERNAL =================== #
# ============================================= #


def _load_config(path: Path, context: dict[str, Any] | None = None) -> Configuration:
    return Configuration.from_file(path, context=context)


def _validate_logging_path(path: str | Path) -> Path:
    # check files exist
    path = Path(path)
    if not path.is_dir():
        raise FileNotFoundError(
            f"Logging directory does not exist or is not a directory: {path.as_posix()}"
        )
    # get configuration file
    config_files = list(path.glob("*.json"))
    if len(config_files) == 0:
        raise FileNotFoundError(
            f"No configuration file found in logging directory: {path.as_posix()}"
        )
    if len(config_files) > 1:
        raise ValueError(
            f"Multiple configuration files found in logging directory: {path.as_posix()}"
        )

    # get "*.log" files from directory
    log_files = list(path.glob("*.log"))
    if len(log_files) == 0:
        raise FileNotFoundError(
            f"No log files found in logging directory: {path.as_posix()}"
        )
    if len(log_files) > 1:
        raise ValueError(
            f"Multiple logs files found in logging directory: {path.as_posix()}"
        )

    return log_files[0], config_files[0]


def _summary(
    log_file: Path,
    config: Configuration,
    output_dir: Path | None = None,
    **kwargs: dict[str, Any],
):
    from .analysis import (
        EventLogParser,
        get_mouse_button_events,
        get_mouse_motion_events,
        get_keyboard_events,
        get_system_monitoring_task_events,
        get_tracking_task_events,
        get_resource_management_task_events,
        get_eyetracking_events,
        get_acceptable_intervals,
        get_unacceptable_intervals,
        get_attention_intervals,
        get_guidance_intervals,
        get_start_and_end_time,
        get_frame_timestamps,
        get_svg_as_image,
    )

    parser = EventLogParser()
    parser.discover_event_classes("matbii")
    events = list(
        parser.parse(log_file, relative_start=kwargs.get("relative_start", True))
    )
    LOGGER.debug(f"Writing output to {output_dir.as_posix()}")
    get_mouse_button_events(parser, events).to_csv(
        output_dir / "mouse_button.csv", index=False
    )
    mouse_motion_df = get_mouse_motion_events(parser, events)
    mouse_motion_df.to_csv(output_dir / "mouse_motion.csv", index=False)
    keyboard_df = get_keyboard_events(parser, events)
    keyboard_df.to_csv(output_dir / "keyboard.csv", index=False)
    eyetracking_df = get_eyetracking_events(parser, events)
    eyetracking_df.to_csv(output_dir / "eyetracking.csv", index=False)
    get_system_monitoring_task_events(parser, events).to_csv(
        output_dir / "system_monitoring.csv", index=False
    )
    get_tracking_task_events(parser, events).to_csv(
        output_dir / "tracking.csv", index=False
    )

    get_resource_management_task_events(parser, events).to_csv(
        output_dir / "resource_management.csv"
    )
    frame_timestamps = pd.DataFrame(get_frame_timestamps(events), columns=["timestamp"])
    frame_timestamps.to_csv(output_dir / "frame_timestamps.csv", index=False)

    # get intervals
    _intervals_as_df(dict(get_acceptable_intervals(events))).to_csv(
        output_dir / "acceptable_intervals.csv", index=False
    )
    _intervals_as_df(dict(get_unacceptable_intervals(events))).to_csv(
        output_dir / "unacceptable_intervals.csv", index=False
    )
    _intervals_as_df(dict(get_guidance_intervals(events))).to_csv(
        output_dir / "guidance_intervals.csv", index=False
    )

    # get attention intervals - mouse, gaze, fixation
    _intervals_as_df(dict(get_attention_intervals(mouse_motion_df))).to_csv(
        output_dir / "attention_intervals_mouse.csv", index=False
    )

    # TODO: get attention intervals - input (mouse + keyboard)

    _intervals_as_df(dict(get_attention_intervals(eyetracking_df))).to_csv(
        output_dir / "attention_intervals_gaze.csv", index=False
    )

    # fixation intervals
    fixation_df = eyetracking_df[eyetracking_df["fixated"]]
    # deal with empty eyetracking df
    if fixation_df.empty:
        columns = list(eyetracking_df.columns)
        columns.remove("fixated")
        fixation_df = pd.DataFrame(columns=columns)
    _intervals_as_df(dict(get_attention_intervals(fixation_df))).to_csv(
        output_dir / "attention_intervals_fixation.csv", index=False
    )

    # record start and end times
    start_time, end_time = get_start_and_end_time(events)

    # make plots
    # output_dir = output_dir / "plots"
    # output_dir.mkdir(parents=True, exist_ok=True)
    attention_mode = config.guidance.attention_mode
    fig = _summary_plot(output_dir, attention_mode=attention_mode)
    fig.savefig(output_dir / "summary.png", bbox_inches="tight")

    # plot eyetracking if we have any

    if not mouse_motion_df.empty:
        img = get_svg_as_image((config.ui.width, config.ui.height), events)
        fig, ax = plt.subplots(figsize=(4, 4))
        fig.suptitle("Mouse motion")
        ax.imshow(img)
        plt.scatter(
            mouse_motion_df["x"],
            mouse_motion_df["y"],
            marker=".",
            alpha=0.5,
            color="red",
        )
        fig.savefig(output_dir / "mouse_motion.png", bbox_inches="tight")

    if not eyetracking_df.empty:
        img = get_svg_as_image((config.ui.width, config.ui.height), events)
        fig, ax = plt.subplots(figsize=(4, 4))
        fig.suptitle("Eyetracking")
        ax.imshow(img)
        plt.scatter(
            eyetracking_df["x"][~eyetracking_df["fixated"]],
            eyetracking_df["y"][~eyetracking_df["fixated"]],
            marker=".",
            alpha=0.2,
            color="blue",
        )
        plt.scatter(
            eyetracking_df["x"][eyetracking_df["fixated"]],
            eyetracking_df["y"][eyetracking_df["fixated"]],
            marker=".",
            alpha=0.8,
            color="red",
        )
        fig.savefig(output_dir / "eyetracking.png", bbox_inches="tight")

    plt.show()


def _summary_plot(
    path: Path,
    attention_mode: Literal["mouse", "fixation", "gaze"] = "mouse",
    guidance_colour: str = "red",
    attention_colour: str = "green",
    system_monitoring_colour: str = SYSTEM_MONITORING_COLOR,
    tracking_colour: str = TRACKING_COLOR,
    resource_management_colour: str = RESOURCE_MANAGEMENT_COLOR,
):
    import matplotlib.pyplot as plt
    from icua.extras.analysis import plot_intervals, plot_timestamps

    fig, ax = plt.subplots(figsize=(20, 2))
    task_data = dict(
        system_monitoring=dict(ylim=(0, 1 / 3), colour=system_monitoring_colour),
        tracking=dict(ylim=(1 / 3, 2 / 3), colour=tracking_colour),
        resource_management=dict(ylim=(2 / 3, 1), colour=resource_management_colour),
    )
    acceptable_intervals = pd.read_csv(path / "acceptable_intervals.csv")
    unacceptable_intervals = pd.read_csv(path / "unacceptable_intervals.csv")
    guidance_intervals = pd.read_csv(path / "guidance_intervals.csv")
    try:
        attention_intervals = pd.read_csv(
            path / f"attention_intervals_{attention_mode}.csv"
        )
    except FileNotFoundError:
        avaliable = [
            p.name.split(".")[0].split("_")[-1]
            for p in path.glob(path / "attention_intervals_*.csv")
        ]
        raise ValueError(
            f"Attention intervals for mode {attention_mode} not found, available: {avaliable}."
        )
    first = True
    for task, data in task_data.items():
        plot_intervals(
            acceptable_intervals[acceptable_intervals["task"] == task],
            color=tracking_colour,
            alpha=0.5,
            label="acceptable" if first else None,
            ymin=data["ylim"][0],
            ymax=data["ylim"][1],
            ax=ax,
        )
        plot_intervals(
            unacceptable_intervals[unacceptable_intervals["task"] == task],
            color="black",
            alpha=0.1,
            label="unacceptable" if first else None,
            ymin=data["ylim"][0],
            ymax=data["ylim"][1],
            ax=ax,
        )
        plot_intervals(
            guidance_intervals[guidance_intervals["task"] == task],
            color=guidance_colour,
            alpha=0.2,
            ax=ax,
            label="guidance" if first else None,
            ymin=data["ylim"][0] + 0.05,
            ymax=data["ylim"][1] - 0.05,
        )
        plot_intervals(
            attention_intervals[attention_intervals["task"] == task],
            color=attention_colour,
            alpha=0.2,
            ax=ax,
            label=f"attention_{attention_mode}" if first else None,
            ymin=data["ylim"][0] + 0.05,
            ymax=data["ylim"][1] - 0.05,
        )

        df = pd.read_csv(path / f"{task}.csv")
        # plot the timestamps for the task changed its state due to the task specific agent.
        plot_timestamps(
            df["timestamp"][~df["user"]],
            color="black",
            alpha=0.5,
            label="task schedule" if first else None,
            ymin=data["ylim"][0] + 0.1,
            ymax=data["ylim"][1] - 0.1,
            ax=ax,
        )
        # plot the timestamps for the task changed its state due to the user.
        plot_timestamps(
            df["timestamp"][df["user"]],
            color="red",
            alpha=0.5,
            label="user input" if first else None,
            ymin=data["ylim"][0] + 0.1,
            ymax=data["ylim"][1] - 0.1,
            ax=ax,
        )
        first = False

    # # plot the start and end times - this is the time of the first and last rendered frame.
    # fig = plot_timestamps(
    #     np.array([start_time, end_time]),
    #     color="black",
    #     ax=ax,
    #     linestyle="--",
    #     label="start/end",
    # )
    plt.legend(loc="lower center", bbox_to_anchor=(0.5, 1), ncol=100)
    plt.tight_layout()
    return fig


def _intervals_as_df(intervals: dict[str, np.ndarray]) -> pd.DataFrame:
    def _gen():
        # yield an empty dataframe to handle the case where no intervals are found
        for task, i in intervals.items():
            if len(i) > 0:
                yield pd.DataFrame({"t1": i[:, 0], "t2": i[:, 1], "task": task})

    result = list(_gen())
    if len(result) == 0:
        return pd.DataFrame(columns=["t1", "t2", "task"])
    df = pd.concat(result)
    df.sort_values(by="t1", inplace=True)
    df.reset_index(drop=True, inplace=True)
    return df



"""Module containing the base class for matbii guidance agents, see `GuidanceAgent` documentation for details."""

from typing import Any
from collections import deque
import time
from icua.agent.actuator_guidance import (
    CounterFactualGuidanceActuator,
)
from icua.event import (
    Event,
    EyeMotionEvent,
    MouseMotionEvent,
)
from icua.agent import GuidanceAgent as _GuidanceAgent
from icua.extras.logging import LogActuator
from icua.utils import LOGGER  # , dict_diff

from star_ray.agent import Actuator, Sensor


class GuidanceAgent(_GuidanceAgent):
    """Base class for matbii guidance agents.

    This class adds some useful fields to the the `beliefs` dict.
    - `is_guidance`: whether guidance is currently being shown for the task.
    - `failure_start`: the time since the last failure started on the task.
    - `guidance_start`: the time since the last guidance started to be shown on the task.

    Other fields are inherited:
    - `is_acceptable`: whether the task is currently acceptable.
    - `is_active`: whether the task is currently active.

    All of these fields are accessible via the corresponding task keys, e.g. `self.beliefs[task]["is_guidance"]`.
    They are updated internally as new observations are received (or actions are taken), it is best not to assign to these fields directly.
    """

    def __init__(  # noqa inherited docs
        self,
        sensors: list[Sensor],
        actuators: list[Actuator],
        counter_factual: bool = False,
        user_input_events: tuple[type[Event]] = None,
        user_input_events_history_size: int | list[int] = 100,
        cycle_times_history_size: int = 10,
    ):
        # this actuator will be used when counter-factual guidance is enabled, any other actuators will be ignored
        _counter_factual_guidance_actuator = CounterFactualGuidanceActuator()
        actuators.append(_counter_factual_guidance_actuator)
        super().__init__(
            sensors, actuators, user_input_events, user_input_events_history_size
        )
        self._counter_factual = counter_factual
        self._counter_factual_guidance_actuator = _counter_factual_guidance_actuator

        # various useful properties used to determine whether guidance should be shown
        self._cycle_times = deque(maxlen=max(cycle_times_history_size, 10))

    def get_cycle_start(self, index: int = 0) -> float:
        """Get the time since the previous cycle started.

        An index of 0 indicates the start of the current cycle.
        If the index is greater than the current history size N (which may happen in the first N - 1 cycles) the oldest cycle time will be returned.

        Args:
            index (int): the index of the cycle to get the start time for.

        Returns:
            float: the time since the given cycle started.
        """
        if index >= self._cycle_times.maxlen:
            raise ValueError(
                f"Invalid argument: `index` {index} must be less than the cycle times history size {self._cycle_times.maxlen}, the can be increased by setting `cycle_times_history_size` in the constructor."
            )
        index = min(index, len(self._cycle_times) - 1)
        return self._cycle_times[index]

    def show_guidance(self, task: str):
        """Show guidance for a given task.

        If overriding you must call super() to ensure consistent behaviour.

        Args:
            task (str): the task to show guidance for.
        """
        self.beliefs[task]["is_guidance"] = True
        self.beliefs[task]["guidance_start"] = self.get_cycle_start()
        if not self._counter_factual:
            for actuator in self.guidance_actuators:
                actuator.show_guidance(task=task)
        else:
            self._counter_factual_guidance_actuator.show_guidance(task=task)

    def hide_guidance(self, task: str):
        """Hide guidance for a given task.

        If overriding you must call super() to ensure consistent behaviour.

        Args:
            task (str): the task(s) to hide guidance for.
        """
        self.beliefs[task]["is_guidance"] = False
        if not self._counter_factual:
            for actuator in self.guidance_actuators:
                actuator.hide_guidance(task=task)
        else:
            self._counter_factual_guidance_actuator.hide_guidance(task=task)

    def log_belief(self, belief: dict[str, Any] | Event) -> None:
        """Method that is intended for logging the beliefs of this agent to a file. This can be very useful for keeping track of the state of the simulation, user input, task acceptability and guidance for post analysis purposes.

        Requires `icua.extras.logging.LogActuator` be attached to this agent and will otherwise have no effect. Logging can be configured in the `LogActuator`, including the way that beliefs are formatted in the file.

        Note that the belief wont be logged immediately, but will be buffered as an action and logged as part of usual action execution cycle (on execute).

        Args:
            belief (dict[str, Any]): belief to log
        """
        actuator: LogActuator | None = next(
            iter(self.get_actuators(oftype=LogActuator)), None
        )
        if actuator:
            actuator.log(belief)
        else:
            LOGGER.warning(
                "Attempted to log a belief without the required actuator: `icua.extras.logging.LogActuator`"
            )

    def __cycle__(self):  # noqa
        # add the latest cycle time (according to this agents cycle)
        self._cycle_times.appendleft(time.time())
        super().__observe__()
        self.decide()
        self.__decide__()

    # def __execute__(self, state: State, *args, **kwargs):  # noqa
    #     # always call this at the end of the cycle (when all beliefs have been updated in subclass)
    #     diff = dict_diff(self._old_beliefs, self.beliefs)
    #     if diff:
    #         # there were some updates to the beliefs, log them and update old beliefs to reflect these changes
    #         self.log_belief(self.beliefs)  # differences in beliefs
    #         self._old_beliefs = deepcopy(self.beliefs)
    #         # print(diff)

    #     # t = time.time()
    #     # print(t - self._timestamp)
    #     # self._timestamp = t
    #     super().__execute__(state, *args, **kwargs)

    def decide(self):
        """Decide method for this agent. This method is intended to determine the guidance actions the agent will take in each cycle. Make use of the `show_guidance` and `hide_guidance` methods to control guidance rather than using raw actions or attempt methods. This will ensure consistent behaviour and handle possible counter-factual guidance."""
        raise NotImplementedError(
            f"Subclasses of {GuidanceAgent.__module__}.{GuidanceAgent.__name__} must implement the `decide` method, this method is intended to determine the guidance actions the agent will take in each cycle."
        )

    # ================================================================================================ #
    # ================================ Below are some useful methods ================================= #
    # ================================================================================================ #

    def time_since_failure_start(self, task: str | None = None) -> float:
        """Get the time since the last failure started on the task (or any task if `task` is None).

        Args:
            task (str | None): the task. If None, use any task.

        Returns:
            float: the time since the last failure started on the task (or any task if `task` is None).
        """
        if task is None:
            failure_time = max(
                self.beliefs[task].get("failure_start", float("nan"))
                for task in self.monitoring_tasks
            )
        else:
            failure_time = self.beliefs[task].get("failure_start", float("nan"))
        return self.get_cycle_start() - failure_time

    def time_since_guidance_start(self, task: str | None = None) -> float:
        """Get the time since the last guidance started to be shown on the task (or any task if `task` is None).

        Args:
            task (str | None): the task. If None, use any task.

        Returns:
            float: the time since the last guidance started to be shown on the task (or any task if `task` is None).
        """
        if task is None:
            guidance_time = max(
                self.beliefs[task].get("guidance_start", float("nan"))
                for task in self.monitoring_tasks
            )
        else:
            guidance_time = self.beliefs[task].get("guidance_start", float("nan"))
        return self.get_cycle_start() - guidance_time

    def on_unacceptable(self, task: str):  # noqa
        self.beliefs[task]["failure_start"] = self.get_cycle_start()
        return super().on_unacceptable(task)

    # ================================================================================================ #
    # =============================== Below are some useful properties =============================== #
    # ================================================================================================ #

    @property
    def guidance_on_tasks(self) -> set[str]:
        """Get the set of tasks that guidance is currently being shown on.

        Returns:
            set[str]: set of tasks that guidance is currently being shown on.
        """
        return set(
            t for t in self.active_tasks if self.beliefs[t].get("is_guidance", False)
        )

    @property
    def fixation_target(self) -> str | None:
        """Get the task that the user is currently fixating on. Eyetracking events arrive quickly, this will check events that have been generated since the last cycle of this agent.

        Returns:
            str | None: the task that the user is currently fixating on, or None if the user is not currently fixating on any task.
        """
        # gather events since the previous cycle
        latest_fixation: EyeMotionEvent | None = None
        # use events from the last 3 cycles, rather than the last cycle... this makes things a bit more robust with timings, 
        # especially if the fixation filter is not great (TODO we could choose different values or make this configurable...?)
        prev_cycle_start = self.get_cycle_start(3)
        _total = 0
        for event in self._user_input_events[EyeMotionEvent]:
            if event.timestamp >= prev_cycle_start:
                _total += 1
                if event.fixated:
                    latest_fixation = event
                    break  # found latest fixation
            else:
                break  # dont check older events
        #print(_total)
        if latest_fixation is None:
            return None
        targets = set(latest_fixation.target) & self.monitoring_tasks
        return next(iter(targets), None)

    @property
    def mouse_target(self) -> str | None:
        """Get the task that the mouse is currently over."""
        event: MouseMotionEvent | None = next(
            self.get_latest_user_input(MouseMotionEvent), None
        )
        if event is None:
            return None
        targets = set(event.target) & self.monitoring_tasks
        return next(iter(targets), None)

    @property
    def gaze_target(self) -> str | None:
        """Get the task that the gaze is currently over."""
        event: EyeMotionEvent | None = next(
            self.get_latest_user_input(EyeMotionEvent), None
        )
        if event is None:
            return None
        targets = set(event.target) & self.monitoring_tasks
        return next(iter(targets), None)



"""Module contains the base class for task acceptability sensors `TaskAcceptabilitySensor`, it is an extension of the base class that is part of `icua` which includes functionality for determining if a task is active based on whether the task element is present in the environment state."""

from typing import Any
from star_ray.event import Event, Observation, ActiveObservation, ErrorObservation
from star_ray_xml import XPathElementsNotFound, Select
from icua.agent import TaskAcceptabilitySensor as _TaskAcceptabilitySensor


class TaskAcceptabilitySensor(_TaskAcceptabilitySensor):
    """This `Sensor` can be used by an agent to track the acceptability of a task."""

    def __init__(self, task_name: str, *args: list[Any], **kwargs: dict[str, Any]):
        """Constructor.

        Args:
            task_name (str): task to track.
            args (list[Any]): Additional optional arguments.
            kwargs (dict[str,Any]): Additional optionals keyword arguments.
        """
        super().__init__(task_name, *args, **kwargs)
        self._is_active = True  # unless it cannot be found...
        # the id of the action that is used to check whether the task is active
        self._is_active_action_id = None

    def is_active(self, task: str = None, **kwargs: dict[str, Any]) -> bool:  # noqa
        return self._is_active  # this is not done by subclass

    def __transduce__(self, observations: list[Observation]) -> list[Observation]:  # noqa
        # fetch the observation that is the result of the is_active action and update _is_active
        # this must happen before beliefs are updated since some updates may depend on whether the task is active
        self._update_is_active(observations)
        return super().__transduce__(observations)

    def on_error_observation(self, observation: ErrorObservation):  # noqa
        if self._is_active:
            return super().on_error_observation(observation)
        # if the task is inactive then some observations may result in XPathElementsNotFound
        if issubclass(observation.resolve_exception_type(), XPathElementsNotFound):
            pass  # ignore these
        else:
            # these may still be relevant
            return super().on_error_observation(observation)

    def _update_is_active(self, observations: list[Observation]) -> bool:
        """Checks whether the task is currently active based on the observation resulting from the `is_active` action."""
        is_active_observation = list(
            filter(
                lambda x: isinstance(x, ActiveObservation)
                and x.action_id == self._is_active_action_id,
                observations,
            )
        )
        if len(is_active_observation) == 0:
            # the is_active observation was not part of these observations
            # this can happen if the sensor is receiving observations via subscription
            return
        # the is_active observation was part of the observations, check it
        is_active_observation = is_active_observation[0]
        if self._is_active:
            if isinstance(is_active_observation, ErrorObservation):
                self._is_active = False  # the task is now inactive!
            else:
                self._is_active = True  # the task is active and remains active
        else:
            if not isinstance(is_active_observation, ErrorObservation):
                self._is_active = True  # the task is inactive and may now be active
            else:
                self._is_active = False  # the task is inactive and remains inactive

    def __sense__(self) -> list[Event]:  # noqa
        if self._is_active:
            actions = self.sense()
            if not isinstance(actions, list | tuple):
                raise TypeError(
                    f"`sense()` must return a `list` of events, received: {type(actions)}"
                )
        else:
            actions = []
        # always check if the task is active
        is_active = Select(xpath=f"//*[@id='{self.task_name}']", attrs=["id"])
        self._is_active_action_id = is_active.id
        actions.insert(0, is_active)
        return actions



"""Package defining agent related functionality."""

from ..tasks import (
    TrackingActuator,
    SystemMonitoringActuator,
    ResourceManagementActuator,
)

__all__ = ("TrackingActuator", "SystemMonitoringActuator", "ResourceManagementActuator")



"""Module contains a guidance sensor for tracking the "system monitoring" task acceptability, see `SystemMonitoringTaskAcceptabilitySensor` documentation for details."""

from typing import Any
from functools import partial
from star_ray_xml import Select
from .sensor_guidance import TaskAcceptabilitySensor

from ..tasks.system_monitoring import SetLightAction, SetSliderAction
from ..utils._const import (
    TASK_ID_SYSTEM_MONITORING,
    slider_id,
    slider_incs_id,
    light_id,
)


class SystemMonitoringTaskAcceptabilitySensor(TaskAcceptabilitySensor):
    """Guidance sensor for the system monitoring task.

    This sensor tracks a number of sub-tasks:

    - `"system_monitoring.light-1"`
    - `"system_monitoring.light-2"`
    - `"system_monitoring.slider-1"`
    - `"system_monitoring.slider-2"`
    - `"system_monitoring.slider-3"`
    - `"system_monitoring.slider-4"`

    The acceptability of these sub-tasks can be checked by calling the methods:

    - `SystemMonitoringTaskAcceptabilitySensor.is_light_acceptable`
    - `SystemMonitoringTaskAcceptabilitySensor.is_slider_acceptable`

    Otherwise follow the `icua.agent.TaskAcceptabilitySensor` API.
    """

    def __init__(self, *args: list[Any], **kwargs: dict[str, Any]):
        """Constructor.

        Args:
            args (list[Any], optional): additional optional arguments.
            kwargs (dict[Any], optional): additional optional keyword arguments.
        """
        super().__init__(TASK_ID_SYSTEM_MONITORING, *args, **kwargs)
        self._is_subtask_acceptable = {
            f"{self.task_name}.light-1": partial(self.is_light_acceptable, 1),
            f"{self.task_name}.light-2": partial(self.is_light_acceptable, 2),
            f"{self.task_name}.slider-1": partial(self.is_slider_acceptable, 1),
            f"{self.task_name}.slider-2": partial(self.is_slider_acceptable, 2),
            f"{self.task_name}.slider-3": partial(self.is_slider_acceptable, 3),
            f"{self.task_name}.slider-4": partial(self.is_slider_acceptable, 4),
        }

    def is_acceptable(self, task: str = None, **kwargs: dict[str, Any]) -> bool:  # noqa
        if not self._is_active:
            return False  # always return False if the task is inactive
        if task is None or task == self.task_name:
            return all([x() for x in self._is_subtask_acceptable.values()])
        else:
            is_acceptable = self._is_subtask_acceptable.get(task, None)
            if is_acceptable is None:
                raise KeyError(
                    f"Invalid subtask: {task}, doesn't exist for task {self.task_name}"
                )
            else:
                return is_acceptable()

    def is_slider_acceptable(self, _id: int) -> bool:
        """Whether the given slider is in an acceptable state.

        Acceptable: the slider is at the central position.
        Unacceptable: otherwise.

        Args:
            _id (int): the id of the slider (1, 2, 3 or 4)

        Returns:
            bool: True if the slider is in an acceptable state, False otherwise.
        """
        if not self._is_active:
            return False  # always return False if the task is inactive
        # sliders should be at the center position to be acceptable
        state = self.beliefs[slider_id(_id)]["data-state"]
        acceptable_state = SetSliderAction.acceptable_state(
            self.beliefs[slider_incs_id(_id)]["incs"]
        )
        return state == acceptable_state

    def is_light_acceptable(self, _id: int) -> bool:
        """Whether the given light is in an acceptable state.

        Acceptable: light-1 is on, light-2 is off.
        Unacceptable: otherwise.

        Args:
            _id (int): the id of the light (1 or 2)

        Returns:
            bool: True if the light is in an acceptable state, False otherwise.
        """
        if not self._is_active:
            return False  # always return False if the task is inactive
        return [self._is_light1_acceptable, self._is_light2_acceptable][_id - 1]()

    def _is_light1_acceptable(self):
        # light 1 should be on
        return self.beliefs[light_id(1)]["data-state"] == SetLightAction.ON

    def _is_light2_acceptable(self):
        # light 2 should be off
        return self.beliefs[light_id(2)]["data-state"] == SetLightAction.OFF

    def sense(self) -> list[Select]:
        """Generates the sense actions that are required for checking whether the system monitoring task is in an acceptable state.

        The actions will request the following data:
        - the state of each light element.
        - the state of each slider element.
        - the number of increments in each slider element.

        Returns:
            list[Select]: list of sense actions to take.
        """
        # take these actions if this task is active
        lights = [f"//*[@id='{light_id(i)}']" for i in (1, 2)]
        sliders = [f"//*[@id='{slider_id(i)}']" for i in (1, 2, 3, 4)]
        slider_incs = [f"//*[@id='{slider_incs_id(i)}']" for i in (1, 2, 3, 4)]

        lights = [Select(xpath=xpath, attrs=["id", "data-state"]) for xpath in lights]
        sliders = [Select(xpath=xpath, attrs=["id", "data-state"]) for xpath in sliders]
        slider_incs = [
            Select(xpath=xpath, attrs=["id", "incs"]) for xpath in slider_incs
        ]
        return [*lights, *sliders, *slider_incs]



"""Matbii package."""

from . import environment
from . import tasks
from . import avatar
from . import agent
from . import guidance
from . import utils
from . import extras

__all__ = (
    "environment",
    "tasks",
    "avatar",
    "agent",
    "guidance",
    "utils",
    "extras",
)



"""Package defining various utilities."""

from ._const import (
    TASK_PATHS,
    TASK_ID_TRACKING,
    TASK_ID_RESOURCE_MANAGEMENT,
    TASK_ID_SYSTEM_MONITORING,
)

from icua.utils import LOGGER
import importlib

__all__ = (
    "LOGGER",
    "get_class_from_fqn",
    "TASK_PATHS",
    "TASK_ID_TRACKING",
    "TASK_ID_RESOURCE_MANAGEMENT",
    "TASK_ID_SYSTEM_MONITORING",
)


def get_class_from_fqn(name: str) -> type:
    """Gets a class from its fully qualified name.

    Returns:
        str : fully qualified name of the class.
    """
    module_path, _, class_name = name.rpartition(".")
    module = importlib.import_module(module_path)
    return getattr(module, class_name)



"""Package defining each of the matbii tasks.

Tasks:
    - tracking
    - system monitoring
    - resource management
"""

from .tracking import AvatarTrackingActuator, TrackingActuator, TargetMoveAction

from .resource_management import (
    ResourceManagementActuator,
    AvatarResourceManagementActuator,
    SetPumpAction,
    BurnFuelAction,
    PumpFuelAction,
    TogglePumpAction,
    TogglePumpFailureAction,
)
from .system_monitoring import (
    SystemMonitoringActuator,
    AvatarSystemMonitoringActuator,
    SetSliderAction,
    SetLightAction,
    ToggleLightAction,
    PerturbSliderAction,
    ResetSliderAction,
)

TASK_EVENT_TYPES = (
    TargetMoveAction,
    SetSliderAction,
    SetLightAction,
    ToggleLightAction,
    SetPumpAction,
    BurnFuelAction,
    PumpFuelAction,
    TogglePumpAction,
    TogglePumpFailureAction,
)


__all__ = (
    "TASK_EVENTS",
    # resource managment
    "ResourceManagementActuator",
    "AvatarResourceManagementActuator",
    "SetPumpAction",
    "TogglePumpAction",
    "BurnFuelAction",
    "PumpFuelAction",
    "TogglePumpFailureAction",
    # tracking
    "AvatarTrackingActuator",
    "TrackingActuator",
    "TargetMoveAction",
    # system monitoring
    "SystemMonitoringActuator",
    "AvatarSystemMonitoringActuator",
    "SetLightAction",
    "ToggleLightAction",
    "SetSliderAction",
    "PerturbSliderAction",
    "ResetSliderAction",
)



"""Entry point for the matbii simulation."""

import importlib.resources
from pathlib import Path
from typing import Any
from collections import defaultdict
import os
import argparse
import importlib
import inspect


from star_ray.utils import _LOGGER

# avoid a pygame issue on linux. TODO this should be moved somewhere more suitable...
os.environ["LD_PRELOAD"] = "/usr/lib/x86_64-linux-gnu/libstdc++.so.6"

# silence logs from star_ray logger
_LOGGER.setLevel("WARNING")


def logging_level(verbosity: int) -> str:
    """Convert a verbosity level to a logging level."""
    return defaultdict(lambda: "DEBUG", {0: "WARNING", 1: "INFO", 2: "DEBUG"})[
        verbosity
    ]


def dot_notation_to_dict(
    dot_notation_dict: dict[str, Any], delim="."
) -> dict[str, Any]:
    """Convert a dictionary with keys in dot notation to a nested dictionary.

    Args:
        dot_notation_dict (dict[str, Any]): the dictionary to convert.
        delim (str, optional): the delimiter used in the dot notation keys. Defaults to ".".

    Returns:
        dict[str, Any]: the nested dictionary.
    """
    nested_dict = {}
    for key, value in dot_notation_dict.items():
        parts = key.split(delim)
        d = nested_dict
        for part in parts[:-1]:
            d = d.setdefault(part, {})
        d[parts[-1]] = value
    return nested_dict


def parse_config_args(unknown_args: list[str]) -> dict[str, Any]:
    """Parse configuration arguments from the unknown arguments."""
    # scan unknown args for those that start with --config.
    # find index of each --config
    from ast import literal_eval

    still_unknown_args = []
    config_args = {}
    config_indices = [i for i, arg in enumerate(unknown_args) if arg.startswith("-")]
    config_indices.append(len(unknown_args))  # need to add the end index
    for i, j in zip(config_indices[:-1], config_indices[1:]):
        if unknown_args[i].startswith("--config"):
            try:
                value = tuple(literal_eval(arg) for arg in unknown_args[i + 1 : j])
            except Exception:
                raise ValueError(
                    f"Error parsing config argument: {unknown_args[i]}, ensure the value(s): {unknown_args[i + 1 : j]} can be interpreted as a python literal. You may need to surround string values with single quotes."
                )
            if len(value) == 1:
                value = value[0]
            config_args[unknown_args[i].replace("--config.", "")] = value
        else:
            still_unknown_args.extend(unknown_args[i:j])
    config_args = dot_notation_to_dict(config_args, delim=".")
    return config_args, still_unknown_args


def parse_cmd_args() -> dict[str, Any]:
    """Parse command line arguments."""
    # load configuration file
    parser = argparse.ArgumentParser()

    # ======================================================================== #
    # =============================== Options ================================ #
    # ======================================================================== #
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity level (use -v, -vv, -vvv for more verbosity)",
    )
    # ======================================================================== #
    # =========================== Options for main =========================== #
    # ======================================================================== #
    parser.add_argument(
        "-c",
        "--config",
        required=False,
        help="Path to configuration file.",
    )
    # ======================================================================== #
    # ========================= Options for examples ========================= #
    # ======================================================================== #
    parser.add_argument(
        "-x",
        "--example",
        required=False,
        help="Name of the example to run - see `matbii.example` or [examples](https://github.com/dicelab-rhul/matbii/tree/main/example) for a full list.",
    )
    # ======================================================================== #
    # ========================= Options for scripts ========================== #
    # ======================================================================== #
    parser.add_argument(
        "-s",
        "--script",
        required=False,
        help="Name of the script to run - see `matbii.extras.scripts`",
    )
    args, unknown_args = parser.parse_known_intermixed_args()
    args = {k.replace("-", "_"): v for k, v in vars(args).items()}
    # create context for configuration from cmd line args, other unknown args are ignored, they may be processed later
    config_args, unknown_args = parse_config_args(unknown_args)
    config_args.setdefault("logging", dict())
    config_args["logging"].setdefault("level", logging_level(args["verbose"]))

    if args.get("example", None) and args.get("script", None):
        raise ValueError("Cannot specify both --example and --script")
    return args, config_args, unknown_args


def run_example(
    name: str, config: str | None = None, config_args: dict[str, Any] | None = None
):
    """Run an example."""
    if config_args is None:
        config_args = dict()

    # navigate to the example directory
    example_dir = Path(importlib.resources.path("matbii", "example"))
    if not example_dir.exists():
        # try again, there are some issues with editable installs
        example_dir = example_dir.parent / "matbii" / example_dir.name
        if not example_dir.exists():
            raise ImportError(f"Example path: {example_dir.as_posix()} not found.")
    example = example_dir / name
    # from matbii.utils import LOGGER
    # LOGGER.debug(f"Using example path: {example.as_posix()}")
    if not example.is_dir():
        avaliable_examples = "\n- ".join(
            [p.name for p in example_dir.iterdir() if p.is_dir()]
        )
        raise ValueError(
            f"Example: `{name}` not found at path: `{example.as_posix()}`. Avaliable examples:\n- {avaliable_examples}"
        )
    # record the current working directory, this will be used to store the log files
    log_dir = Path.cwd() / f"{name}-logs"
    config_args.setdefault("logging", dict())
    config_args["logging"]["path"] = log_dir.as_posix()
    config_args.setdefault("experiment", dict())
    config_args["experiment"]["path"] = example.as_posix()

    # TODO we might search for this instead so that the file name doesnt need to be experiment.json
    if config:
        config_file = example / config
    else:
        config_file = example / "experiment.json"
    if not config_file.exists():
        raise FileNotFoundError(
            f"Example config file: {config_file.as_posix()} not found"
        )
    from matbii.utils import LOGGER

    LOGGER.debug(f"Example path: {example.as_posix()}")
    from matbii.main import main

    main(config_file.as_posix(), **config_args)


def run_script(name: str, **kwargs: dict[str, Any]):
    """Run a script."""
    from matbii.extras import scripts

    avaliable_scripts = {
        func[0]: func[1]
        for func in inspect.getmembers(scripts, inspect.isfunction)
        if not func[0].startswith("_")  # Filter out private functions
    }
    if name not in avaliable_scripts:
        script_names = "\n- ".join(avaliable_scripts.keys())
        raise ValueError(
            f"Script: `{name}` not found, avaliable scripts:\n- {script_names}"
        )
    avaliable_scripts[name](**kwargs)


def _unknown_args_error(unknown_args: Any):
    if unknown_args:
        raise ValueError(
            f"Unknown arguments: {unknown_args}, see `matbii --help` for options."
        )


def main():
    """Main entry point for the matbii."""
    args, config_args, unknown_args = parse_cmd_args()

    # premeptively set the logging level
    from matbii.utils import LOGGER

    LOGGER.set_level(config_args["logging"].get("level", "WARNING"))

    # 1. Run a script if --script is specified
    if args.get("script", None):
        # unknown arguments are ok, they will be grabbed by the script
        run_script(args.get("script"), **args)
        return

    _unknown_args_error(unknown_args)
    # 2. Run an example if --example is specified
    if args.get("example", None):
        run_example(
            args.get("example"),
            config=args.get("config", None),
            config_args=config_args,
        )
        return

    # 3. Run the main simulation
    if args.get("config", None):
        from matbii.main import main

        main(**args)
    else:
        raise ValueError(
            "No configuration provided, use -c or --config to specify a config file."
        )


if __name__ == "__main__":
    main()



"""Implementation of an actuator that will close the window when the escape key is pressed."""

from icua.agent import Actuator, attempt
from icua.event import KeyEvent, WindowCloseEvent
from ..utils import LOGGER


class ExitActuator(Actuator):
    """AvatarActuator for the MATBII avatar."""

    @attempt
    def exit(self, action: KeyEvent):
        """Attempt method for a `KeyEvent` that will convert a press to `Esc` to a `WindowCloseEvent`, triggering the environment to terminate.

        Args:
            action (KeyEvent): user input event.

        Returns:
            WindowCloseEvent: event to close the window.
        """
        if isinstance(action, KeyEvent):
            if action.key == "escape":
                LOGGER.info("Escape was pressed, shutting down...")
                return WindowCloseEvent()



"""Module that contains useful post analysis functions."""

from .get_task import (
    get_system_monitoring_task_events,
    get_resource_management_task_events,
    get_tracking_task_events,
)

from icua.extras.analysis import (
    EventLogParser,
    get_mouse_button_events,
    get_mouse_motion_events,
    get_keyboard_events,
    get_eyetracking_events,
    get_acceptable_intervals,
    get_unacceptable_intervals,
    get_guidance_intervals,
    get_attention_intervals,
    get_start_and_end_time,
    get_svg_as_image,
    get_frame_timestamps,
    merge_intervals,
    isin_intervals,
)

__all__ = [
    "EventLogParser",
    "get_system_monitoring_task_events",
    "get_resource_management_task_events",
    "get_tracking_task_events",
    "get_mouse_button_events",
    "get_mouse_motion_events",
    "get_keyboard_events",
    "get_eyetracking_events",
    "get_acceptable_intervals",
    "get_unacceptable_intervals",
    "get_guidance_intervals",
    "get_attention_intervals",
    "get_start_and_end_time",
    "get_frame_timestamps",
    "get_svg_as_image",
    "merge_intervals",
    "isin_intervals",
]



"""Package that defines the matbii tracking task."""

from .tracking import AvatarTrackingActuator, TrackingActuator, TargetMoveAction

__all__ = ("AvatarTrackingActuator", "TrackingActuator", "TargetMoveAction")



"""Package that defines the matbii system monitoring task."""

from .system_monitoring import (
    SystemMonitoringActuator,
    AvatarSystemMonitoringActuator,
    SetSliderAction,
    SetLightAction,
    ToggleLightAction,
    PerturbSliderAction,
    ResetSliderAction,
)


__all__ = (
    "SystemMonitoringActuator",
    "AvatarSystemMonitoringActuator",
    "SetLightAction",
    "ToggleLightAction",
    "SetSliderAction",
    "PerturbSliderAction",
    "ResetSliderAction",
)



"""Tests for the class: `matbii.config.Configuration`."""

from matbii.config import Configuration
from matbii.utils import get_class_from_fqn
from matbii.guidance import ArrowGuidanceActuator


def test_configuration():
    """Tests that the validators in `Configuration` are working correctly on default data."""
    Configuration()


def test_get_class_from_fqn():
    """Tests the utility function `get_class_from_fqn`."""
    fqn = "matbii.guidance.ArrowGuidanceActuator"
    cls = get_class_from_fqn(fqn)
    if cls != ArrowGuidanceActuator:
        raise ValueError(f"Failed to find class by fully qualified name: {fqn}.")



"""Package defining avatar related functionality."""

# avatar
from icua.agent import Avatar, AvatarActuator
from .exit_actuator import ExitActuator
from ..tasks import (
    AvatarTrackingActuator,
    AvatarSystemMonitoringActuator,
    AvatarResourceManagementActuator,
)

# eyetracking
from icua.extras.eyetracking import EyetrackerIOSensor, tobii

__all__ = (
    "Avatar",
    "AvatarActuator",
    "ExitActuator",
    "AvatarTrackingActuator",
    "AvatarSystemMonitoringActuator",
    "AvatarResourceManagementActuator",
    "WindowConfiguration",
    "EyetrackerIOSensor",
    "tobii",
)
